from typing import List

import sys

from contextlib import contextmanager
from pathlib import Path

import click
import mckit.version as meta

from mckit.cli.commands import do_check, do_compose, do_decompose, do_split, do_transform
from mckit.cli.commands.common import get_default_output_directory
from mckit.cli.logging import init_logger, logger
from mckit.utils import MCNP_ENCODING

NAME = meta.__title__
VERSION = meta.__version__
LOG_FILE_RETENTION = 3
NO_LEVEL_BELOW = 30


context = {}


@click.group(help=meta.__summary__)
@click.option("--override/--no-override", default=False)
@click.option("--verbose/--no-verbose", default=False, help="Log everything")
@click.option("--quiet/--no-quiet", default=False, help="Log only WARNINGS and above")
@click.option("--logfile", default=None, help="File to log to")
@click.version_option(VERSION, prog_name=NAME)
def mckit(verbose: bool, quiet: bool, logfile: str, override: bool) -> None:
    # """MCKIT command line utility."""
    init_logger(logfile, quiet, verbose)
    #
    # TODO dvp: add customized logger configuring from a configuration toml-file.
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below
    # obj = ctx.ensure_object(dict)
    # obj["DEBUG"] = debug
    context["OVERRIDE"] = override


@mckit.command()
@click.option("--output", "-o", default=None, help="Output directory, default: <source>.universes")
@click.option(
    "--fill-descriptor",
    "-f",
    default="fill-descriptor.toml",
    help="TOML file for FILL descriptors",
)
@click.argument("source", metavar="<source>", type=click.Path(exists=True), nargs=1, required=True)
def decompose(output, fill_descriptor, source):
    """Separate an MCNP model to envelopes and filling universes."""
    return do_decompose(output, fill_descriptor, source, context["OVERRIDE"])


@mckit.command()
@click.option("--output", "-o", required=True, help="Output file")
@click.option("--fill-descriptor", default=None, help="TOML file for FILL descriptors")
@click.argument("source", metavar="<source>", type=click.Path(exists=True), nargs=1, required=True)
def compose(output, fill_descriptor, source):
    """Merge universes and envelopes into MCNP model using merge descriptor."""
    if fill_descriptor is None:
        fill_descriptor = Path(source).absolute().parent / "fill-descriptor.toml"
    else:
        fill_descriptor = Path(fill_descriptor)
    if not fill_descriptor.exists():
        raise click.UsageError(f'Cannot find fill descriptor file "{fill_descriptor}"')
    logger.info(
        'Composing "{output}", from envelopes "{source}" with fill descriptor "{fill_descriptor}"',
        output=output,
        source=source,
        fill_descriptor=fill_descriptor,
    )
    return do_compose(output, fill_descriptor, source, context["OVERRIDE"])


@mckit.command()
@click.option("--output", "-o", default=None, help="Output directory")
@click.option(
    "--separators/--no-separators",
    default=False,
    help="Write files with decorative comments to separate this model sections (cells, surfaces etc.) on concatenation",
)
@click.argument("source", metavar="<source>", type=click.Path(exists=True), nargs=1, required=True)
def split(output, source, separators):
    """Splits MCNP model to text portions (opposite to concat)."""
    if output is None:
        output = get_default_output_directory(source, ".split")
    else:
        output = Path(output)
    output.mkdir(parents=True, exist_ok=True)
    logger.info("Running mckit split")
    logger.debug("Working dir {}", Path(".").absolute())
    logger.info('Splitting "{source}" to directory "{output}"', source=source, output=output)
    return do_split(output, source, context["OVERRIDE"], separators)


# noinspection PyCompatibility
@contextmanager
def resolve_output(output, exist_ok=False, encoding=MCNP_ENCODING):
    if output:
        output = Path(output)
        if exist_ok or not output.exists():
            fid = output.open("w", encoding=encoding)
        else:
            raise click.UsageError(
                f"Specify --override option to override existing file '{output}'"
            )
    else:
        fid = sys.stdout
    try:
        yield fid
    finally:
        if fid is not sys.stdout:
            fid.close()


# noinspection PyCompatibility
@mckit.command()
@click.option(
    "--output",
    "-o",
    metavar="<output>",
    type=click.Path(exists=False),
    required=False,
    help="File to write the concatenated parts",
)
@click.option(
    "--parts-encoding",
    metavar="<output>",
    type=click.Path(exists=False),
    required=False,
    default=MCNP_ENCODING,
    help=f"Encoding to read parts (default:{MCNP_ENCODING})",
)
@click.option(
    "--output-encoding",
    metavar="<output>",
    type=click.Path(exists=False),
    required=False,
    default=MCNP_ENCODING,
    help=f"Encoding to write output (default:{MCNP_ENCODING})",
)
@click.argument("parts", metavar="<part...>", type=click.Path(exists=True), nargs=-1, required=True)
def concat(output, parts_encoding, output_encoding, parts):
    """Concat text files.

    (will filter texts according specification in future)
    """
    override = context["OVERRIDE"]
    with resolve_output(output, exist_ok=override, encoding=output_encoding) as out_fid:
        for f in parts:
            f = Path(f)
            # TODO dvp: Add filtering of a part's text here. Implement as external scripts call.
            #           Should be configurable
            print(f.read_text(encoding=parts_encoding), file=out_fid, end="")


# noinspection PyCompatibility
@mckit.command()
@click.argument(
    "sources", metavar="<source>", type=click.Path(exists=True), nargs=-1, required=True
)
def check(sources: List[click.Path]) -> None:
    """Read MCNP model(s) and show statistics and clashes."""
    for source in sources:
        do_check(source)


# noinspection PyCompatibility
@mckit.command()
@click.option(
    "--transformation",
    "-t",
    type=click.INT,
    required=True,
    help="Number of transformation to apply",
)
@click.option(
    "--output",
    "-o",
    type=click.STRING,
    required=True,
    help="Output file",
)
@click.option(
    "--transformations",
    "-i",
    default="transformations.txt",
    type=click.Path(exists=True),
    help="Transformations specification file (default: transformations.txt)",
)
@click.argument("source", metavar="<source>", type=click.Path(exists=True), nargs=1, required=True)
def transform(
    output: click.STRING,
    transformation: int,
    transformations: click.Path,
    source: click.Path,
) -> None:
    """Transform MCNP model(s) with one of specified transformation."""
    do_transform(
        Path(output),
        transformation,
        Path(str(transformations)),
        Path(str(source)),
        context["OVERRIDE"],
    )
    logger.info("File {} is transformed to {}", source, output)


if __name__ == "__main__":
    mckit(obj={})



import pytest

from mckit.source import Distribution, Source


class TestDistribution:
    @pytest.mark.parametrize(
        "name, values, probs",
        [
            (1, [1, 2, 3], [1, 2, 3, 4, 5]),
            (2, [1, 2, 3], [1, 2, 3, 4]),
            (3, [1, 2, 3, 4], [1, 2]),
            (4, [1, 2, 3, 4], Distribution(1, [1, 2], [1, 2])),
        ],
    )
    def test_create_failure(self, name, values, probs):
        with pytest.raises(ValueError):
            Distribution(name, values, probs)

    @pytest.mark.parametrize(
        "name, values, probs, var",
        [
            (1, [1, 2, 3], [1, 2, 3], "X"),
            (2, [1, 2, 3, 4], [1, 2, 3], "Y"),
            (3, [[0.1, 0.5, 0.4], [1, 0, 0]], [0.3, 0.7], "VEC"),
        ],
    )
    def test_create(self, name, values, probs, var):
        d = Distribution(name, values, probs, variable=var)
        assert d.name == name
        assert d.size == len(probs)
        assert len(d) == len(probs)
        assert d.variable == var

    @pytest.mark.parametrize(
        "name, values, probs, expected_names",
        [
            (1, [1, 2, 3], [1, 2, 3], set()),
            (
                2,
                [Distribution(10, [1, 2], [1, 2]), Distribution(11, [1, 2], [2])],
                [1, 2],
                {10, 11},
            ),
            (
                3,
                [Distribution(10, [1, 2], [1, 2]), Distribution(11, [1, 2], [2])],
                Distribution(12, [1, 2], [1, 2]),
                {10, 11},
            ),
            (1, [1, 2, 3], Distribution(12, [1, 2], [1, 2]), set()),
        ],
    )
    def test_get_inner(self, name, values, probs, expected_names):
        d = Distribution(name, values, probs)
        inner = d.get_inner()
        inner_names = {i.name for i in inner}
        assert inner_names == expected_names

    @pytest.mark.parametrize(
        "name, values, probs, discrete, expected",
        [
            (1, [1, 2, 3], [4, 5, 6], True, "SI1 L 1 2 3\nSP1 D 4 5 6"),
            (2, [1, 2, 3], [4, 5], False, "SI2 H 1 2 3\nSP2 D 0 4 5"),
            (3, [1, 2, 3], Distribution(4, [4, 5, 6], [7, 8, 9]), True, "DS3 L 1 2 3"),
            (
                4,
                [Distribution(10, [1, 2], [1, 2]), Distribution(11, [1, 2], [2])],
                [1, 2],
                False,
                "SI4 S 10 11\nSP4 D 1 2",
            ),
            (
                5,
                [Distribution(10, [1, 2], [1, 2]), Distribution(11, [1, 2], [2])],
                Distribution(12, [1, 2], [1, 2]),
                False,
                "DS5 S 10 11",
            ),
            (
                4,
                [[0.5, 0.3, -0.2], [-0.9, 0, 0.4]],
                [0.3, 0.7],
                True,
                "SI4 L 0.5 0.3 -0.2 -0.9 0 0.4\nSP4 D 0.3 0.7",
            ),
            (
                5,
                [[0, -1, 0], [0.5, 0, 0.3]],
                Distribution(4, [5, 6], [7, 8]),
                True,
                "DS5 L 0 -1 0 0.5 0 0.3",
            ),
            (6, [1, 2, 3], [4, 5, 6], False, "SI6 A 1 2 3\nSP6 4 5 6"),
        ],
    )
    def test_mcnp_repr(self, name, values, probs, discrete, expected):
        d = Distribution(name, values, probs, is_discrete=discrete)
        assert d.mcnp_repr() == expected

    @pytest.mark.parametrize(
        "name, values, probs, expected",
        [
            (1, [1, 2, 3], [4, 5, 6], None),
            (2, [1, 2, 3], [4, 5], None),
            (3, [1, 2, 3], Distribution(4, [4, 5, 6], [7, 8, 9]), 0),
            (
                4,
                [Distribution(10, [1, 2], [1, 2]), Distribution(11, [1, 2], [2])],
                [1, 2],
                None,
            ),
            (
                5,
                [Distribution(10, [1, 2], [1, 2]), Distribution(11, [1, 2], [2])],
                Distribution(12, [1, 2], [1, 2]),
                0,
            ),
        ],
    )
    def test_depends_on(self, name, values, probs, expected):
        d = Distribution(name, values, probs)
        if expected is None:
            assert d.depends_on() is None
        else:
            assert d.depends_on() is probs


class TestSource:
    distrs = [
        Distribution(1, [1, 2], [1]),
        Distribution(2, [2, 3], [1]),
        Distribution(3, [4, 5, 6], [1, 2, 3], "X", True),
    ]

    @pytest.mark.parametrize(
        "kwargs, expected",
        [
            ({"PAR": 1, "ERG": 14}, "SDEF PAR=1 ERG=14"),
            ({"PAR": 1, "X": distrs[0]}, "SDEF PAR=1 X=D1\nSI1 H 1 2\nSP1 D 0 1"),
            (
                {"PAR": 2, "X": distrs[0], "Y": distrs[2]},
                "SDEF PAR=2 X=D1 Y=D3\nSI1 H 1 2\nSP1 D 0 1\nSI3 L 4 5 6\nSP3 D 1 2 3",
            ),
            (
                {
                    "PAR": 2,
                    "X": distrs[2],
                    "Y": Distribution(2, [1, 2, 3], distrs[2], is_discrete=True),
                },
                "SDEF PAR=2 X=D3 Y=FX D2\nSI3 L 4 5 6\nSP3 D 1 2 3\nDS2 L 1 2 3",
            ),
            (
                {
                    "PAR": 3,
                    "X": distrs[2],
                    "Y": Distribution(
                        4,
                        [
                            Distribution(10, [2, 3], [1]),
                            Distribution(11, [3, 4], [1]),
                            Distribution(12, [4, 5], [1]),
                        ],
                        distrs[2],
                    ),
                    "Z": Distribution(
                        5,
                        [
                            Distribution(13, [6, 7], [1]),
                            Distribution(14, [7, 8], [1]),
                            Distribution(15, [8, 9], [1]),
                        ],
                        distrs[2],
                    ),
                },
                "SDEF PAR=3 X=D3 Y=FX D4 Z=FX D5\nSI3 L 4 5 6\nSP3 D 1 2 3\n"
                "DS4 S 10 11 12\nDS5 S 13 14 15\nSI10 H 2 3\nSP10 D 0 1\n"
                "SI11 H 3 4\nSP11 D 0 1\nSI12 H 4 5\nSP12 D 0 1\n"
                "SI13 H 6 7\nSP13 D 0 1\nSI14 H 7 8\nSP14 D 0 1\n"
                "SI15 H 8 9\nSP15 D 0 1",
            ),
        ],
    )
    def test_mcnp_repr(self, kwargs, expected):
        src = Source(**kwargs)
        assert src.mcnp_repr() == expected



"""Build mckit.

The `build` method from this module is called by poetry build system
to set up proper options for actual builder.

The module builds and arrange C-dependency ``nlopt`` before own build.
"""
from typing import Any, Dict, List, Optional

# import distutils.log as log
import shutil

from pathlib import Path
from pprint import pprint

from building.build_nlopt import build_nlopt
from building.extension_geometry import GeometryExtension
from building.extension_utils import MACOS, WIN, get_library_dir
from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools.dist import Distribution

DIR = Path(__file__).parent


class BinaryDistribution(Distribution):
    def is_pure(self):
        return False


def get_nlopt_lib_name() -> str:
    """Compute library name: this depends on OS and python version."""
    if MACOS:
        return f"libnlopt.dylib"
    if WIN:
        return f"Release/nlopt.dll"
    return f"libnlopt.so"


class MCKitBuilder(build_ext):
    def __init__(self, dist: Distribution, **kwargs) -> None:
        build_ext.__init__(self, dist, **kwargs)

    def finalize_options(self):
        build_ext.finalize_options(self)
        # Late import to use numpy installed on isolated build
        import numpy as np

        library_dir = get_library_dir()
        self.library_dirs.append(str(library_dir))
        self.include_dirs.append(np.get_include())
        self.include_dirs.append(str(library_dir.parent / "include"))

    def build_extension(self, extension: Extension) -> None:
        nlopt_build_dir = build_nlopt(clean=True)
        nlopt_lib = nlopt_build_dir / get_nlopt_lib_name()
        # log.info(f"---***  builder.build_lib: {self.build_lib}")
        # log.info(f"---***  builder.include_dirs: {self.include_dirs}")
        # log.info(f"---***  builder.library_dirs: {self.library_dirs}")
        # log.info(f"---***  nlopt lib path: {nlopt_lib}")
        ext_dir = Path(self.get_ext_fullpath(extension.name)).parent.absolute()
        # log.info(f"---***  copy nlopt lib to {ext_dir}")
        if ext_dir.exists():
            save_library(ext_dir, nlopt_lib)
        save_library(DIR / "src/mckit", nlopt_lib)
        # log.info("---*** Defined geometry extension:")
        # log.info(str(extension))
        # log.info("---***")
        build_ext.build_extension(self, extension)
        # log.info("---*** Search geometry:")
        # log.info(list(DIR.glob("**/*geometry*")))


def build(setup_kwargs: Dict[str, Any]) -> None:
    """Set specific distribution options.

    This function is called with setup.py generated by pip from pyproject.toml.
    """
    update_package_data(setup_kwargs)
    update_setup_requires(setup_kwargs)
    geometry_extension = GeometryExtension()
    setup_kwargs.update(
        {
            "ext_modules": [geometry_extension],
            "cmdclass": {"build_ext": MCKitBuilder},
            "distclass": BinaryDistribution,
            "long_description": Path("README.rst").read_text(encoding="utf8"),
            "src_root": str(Path(__file__).parent),
        }
    )
    save_setup_kwargs(setup_kwargs)


def update_setup_requires(setup_kwargs: Dict[str, Any]) -> None:
    """Fix for poetry issue: it doesn't install setup requirements."""
    setup_requires = setup_kwargs.get("setup_requires")  # type: Optional[List[str]]
    assert setup_requires is None, "Poetry has created setup_requires! Check the setup-generated.py"
    setup_requires = [
        "poetry-core>=1.0.7",
        "setuptools>=58.1",
        "wheel",
        # "cmake>=3.18.4",
        # "numpy>=1.13",
        # "mkl-devel",
    ]
    setup_kwargs["setup_requires"] = setup_requires
    save_generated_setup()


def update_package_data(setup_kwargs: Dict[str, Any]) -> None:
    """Fix for poetry issue: it doesn't provide correct specification from `[tool.poetry].input` field."""
    package_data = [
        "data/isotopes.dat",
        "nlopt.dll" if WIN else "libnlopt*",
    ]
    setup_kwargs["package_data"] = {"mckit": package_data}


def save_library(destination: Path, so: Path) -> None:
    if not so.exists():
        raise FileNotFoundError(f"Cannot find shared library {so}")
    shutil.copy(str(so), str(destination))


def save_setup_kwargs(setup_kwargs: Dict[str, Any]) -> None:
    """Save resulting setup_kwargs for examining."""
    kwargs_path = Path(__file__).parent / "poetry_setup_kwargs.txt"
    with kwargs_path.open(mode="w") as fid:
        pprint(setup_kwargs, fid, indent=4)


def save_generated_setup() -> None:
    """Save generated setup.py.

    Poetry ignores MANIFEST.in on command: ::

        poetry build -f sdist.

    So, 3rd-party code and build scripts are not included. Besides, output format is just default one.

    However, command: ::

        python setup-generated.py sdist --formats=gztar,xztar,zip

    uses MANIFEST.in and creates proper sdist archives.
    This is used in GitHub Action "Release" (see `.github/workflows/release.yml`).
    Besides, the setup-generated.py script can be used for debugging of setup process.

        Note: ::

        This file is regenerated on every build, so, there's no reason to store it in Git.
    """
    my_dir = Path(__file__).parent
    src = my_dir / "setup.py"
    dst = my_dir / "setup-generated.py"
    if src.exists():
        shutil.copy(str(src), str(dst))



import mckit.parser.common.utils as cmn
import sly

from mckit.parser.common import Lexer as LexerBase
from mckit.parser.common.utils import drop_c_comments, extract_comments
from mckit.transformation import Transformation


# noinspection PyPep8Naming,PyUnboundLocalVariable,PyUnresolvedReferences,SpellCheckingInspection
class Lexer(LexerBase):
    tokens = {NAME, FLOAT, INTEGER}

    NAME = r"\s{0,5}\*?tr\d+"

    @_(r"\s{0,5}\*?tr\d+")
    def NAME(self, t):
        if t.value[0].isspace():
            t.value = t.value.lstrip()
        if t.value[0] == "*":
            in_degrees = True
            name = int(t.value[3:])
        else:
            in_degrees = False
            name = int(t.value[2:])
        t.value = name, in_degrees
        return t

    @_(cmn.FLOAT)
    def FLOAT(self, t):
        return LexerBase.on_float(t, use_zero=False)

    @_(cmn.INTEGER)
    def INTEGER(self, t):
        return LexerBase.on_integer(t, use_zero=False)


# noinspection PyUnresolvedReferences
class Parser(sly.Parser):
    tokens = Lexer.tokens

    @_("NAME transform_params")
    def transformation(self, p):
        name, in_degrees = p.NAME
        translation, rotation, inverted = p.transform_params
        return Transformation(
            translation=translation,
            rotation=rotation,
            indegrees=in_degrees,
            inverted=inverted,
            name=name,
        )

    @_("translation rotation")
    def transform_params(self, p):
        rotation, inverted = p.rotation
        return p.translation, rotation, inverted

    @_("translation")
    def transform_params(self, p):
        return p.translation, None, False

    @_("float float float")
    def translation(self, p):
        return [f for f in p]

    # TODO dvp: check what to do, if transformation is specified with default values using the MCNP J shortcuts?

    @_("float float float float float float float float float INTEGER")
    def rotation(self, p):
        m = p[9]
        assert m == -1 or m == 1, f"Invalid M option value {m}"
        return [f for f in p][:-1], m == -1

    @_(
        "float float float float float float float float float",
        "float float float float float float",
        "float float float float float",
        "float float float",
    )
    def rotation(self, p):
        return [f for f in p], False

    @_("FLOAT")
    def float(self, p):
        return p.FLOAT

    @_("INTEGER")
    def float(self, p):
        return float(p.INTEGER)


def parse(text: str) -> Transformation:
    text = drop_c_comments(text)
    text, comments, trailing_comments = extract_comments(text)
    lexer = Lexer()
    parser = Parser()
    result: Transformation = parser.parse(lexer.tokenize(text))
    if trailing_comments:
        result.options["comment"] = trailing_comments
    return result



import numpy as np
import pytest

from mckit.utils.tolerance import DEFAULT_TOLERANCE_ESTIMATOR

SOME_ARRAY = np.arange(3, dtype=float)


@pytest.mark.parametrize(
    "a, b, expected, msg",
    [
        (1.0, 1.0 + 1.0e-12, False, "#1 1.0e-12 is to be distinguished"),
        (1.0, 1.0 + 1.0e-13, True, "#2 1.0e-13 is not to be distinguished"),
        (
            SOME_ARRAY,
            SOME_ARRAY + 1.0e-11,
            False,
            "#3 on ndarray 1.0e-11 is to be distinguished",
        ),
        (
            SOME_ARRAY,
            SOME_ARRAY + 1.0e-13,
            True,
            "#4 on ndarray 1.0e-12 is not to be distinguished",
        ),
        ((1.0, SOME_ARRAY), (1.0, SOME_ARRAY), True, "#5 iterables"),
        (1, 1, True, "#6 Integers"),
    ],
)
def test_tolerance_estimator(a, b, expected, msg):
    actual = DEFAULT_TOLERANCE_ESTIMATOR(a, b)
    assert actual == expected, msg


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-

import mckit.parser.material_parser as mp
import pytest

from mckit.material import Composition, Element


@pytest.mark.parametrize(
    "text, expected_types, expected_values",
    [
        ("1001.21c -1.0", ["FRACTION"], [(1001, "21c", -1.0)]),
        ("  M100 1000 1.0", ["NAME", "FRACTION"], [100, (1000, None, 1.0)]),
        (
            "  M100 1000 1.0 gas=1",
            ["NAME", "FRACTION", "OPTION"],
            [100, (1000, None, 1.0), "gas=1"],
        ),
        ("M17    18000.70c 1.", ["NAME", "FRACTION"], [17, (18000, "70c", 1.0)]),
    ],
)
def test_composition_lexer(text, expected_types, expected_values):
    lexer = mp.Lexer()
    tokens = list(lexer.tokenize(text))
    result = list(t.type for t in tokens)
    assert result == expected_types
    result = list(t.value for t in tokens)
    assert result == expected_values


# M17    18000.70c 1. $Ar 00 weight(%)  100


@pytest.mark.parametrize(
    "text, expected",
    [
        (
            """m1
        1001.21c -1.0
        """,
            Composition(weight=[(1001, 1.0)], name=1),
        ),
        (
            """m1000
        1001.21c -1.0 $ eol comment
        """,
            Composition(
                weight=[
                    (
                        Element(
                            1001,
                            lib="21c",
                            # comment="$ eol comment"
                        ),
                        1.0,
                    )
                ],
                name=1000,
            ),
        ),
        (
            """M1000
        1001.21c -1.0
            $ trailing comment1
            $ trailing comment2
        """,
            Composition(
                weight=[(Element(1001, lib="21c"), 1.0)],
                name=1000,
                # comment=["trailing comment1", "trailing comment2"],
            ),
        ),
        (
            """M1000
    1001.21c -1.0
c something
        $ trailing comment1
        $ trailing comment2
        """,
            Composition(
                weight=[(Element(1001, lib="21c"), 1.0)],
                name=1000,
                # comment=["trailing comment1", "trailing comment2"],
            ),
        ),
        (
            """M1000
        1001.21c -1.0
            gas 1
            $ trailing comment1
            $ trailing comment2
        """,
            Composition(
                weight=[(Element(1001, lib="21c"), 1.0)],
                name=1000,
                # comment=["trailing comment1", "trailing comment2"],
                GAS=1,
            ),
        ),
        (
            "m1 1001 0.1 1002 0.9",
            Composition(atomic=[(1001, 0.1), (1002, 0.9)], name=1),
        ),
        (
            "m3 1001 0.1 1002 0.9 gas=1",
            Composition(atomic=[(1001, 0.1), (1002, 0.9)], name=3, GAS=1),
        ),
    ],
)
def test_composition_parser(text, expected):
    result = mp.parse(text)
    assert isinstance(result, Composition), "Parser should create instance of Composition"
    assert result == expected
    assert result.options == expected.options


if __name__ == "__main__":
    pytest.main()



"""Code to build mckit geometry C-extension."""
from __future__ import annotations

from typing import Iterable, List

import os
import shlex
import sys

from pathlib import Path

from building.extension_utils import MACOS, WIN, get_library_dir
from setuptools import Extension as _Extension

# See MKL linking options for various versions of MKL and OS:
# https://software.intel.com/content/www/us/en/develop/tools/oneapi/components/onemkl/link-line-advisor.html
#
# Windows (oneMKL 2021, MS-C/C++,64bit, c-32bit, shared library, no cluster):
# mkl_intel_lp64_dll.lib mkl_tbb_thread_dll.lib mkl_core_dll.lib tbb.lib

if not WIN:

    def _make_full_names(lib_dir: Path, mkl_libs: Iterable[str]) -> List[str]:
        """Add library directory and extension to an MKL library name.

        According to the recommendation
        https://community.intel.com/t5/Intel-oneAPI-Math-Kernel-Library/pypi-package-for-mkl-2021-2-0-is-not-linkable-under-Linux-and/m-p/1275110?profile.language=ru
        on Linux MKL requires full path to the library.
        This method construct appropriate paths and searches the libraries.

        Args:
            lib_dir: directory, where to search libraries
            mkl_libs: list of library names to find

        Returns:
             The list of found libraries.

        Raises:
            EnvironmentError: if a library is not found.
        """
        # TODO dvp: implement logic for other possible suffixes in future MKL versions
        if MACOS:
            suffix = "2.dylib"
        else:
            if sys.platform != "linux":
                msg = f"Unknown platform {sys.platform}"
                raise EnvironmentError(msg)
            suffix = "so.2"

        lib_paths: list[Path] = [lib_dir / f"lib{_p}.{suffix}" for _p in mkl_libs]

        for p in lib_paths:
            if not p.exists():
                msg = f"{p} is not a valid path to an MKL library."
                raise EnvironmentError(msg)

        return [str(_p) for _p in lib_paths]


class GeometryExtension(_Extension):
    """The :class:`_Extension` specialization for mckit geometry module."""

    def __init__(self) -> None:

        super().__init__(
            "mckit.geometry",
            sources=list(map(str, Path("src", "mckit", "src").glob("*.c"))),
            language="c",
        )

        if WIN:
            cflags = ["/EHsc", "/bigobj"]
            ldflags = []
            libraries = [
                "mkl_rt",
                "nlopt",
            ]
        else:
            cflags = [
                "-O3",
                "-Wall",
                "-m64",
            ]
            env_cflags = os.environ.get("CFLAGS", "")
            env_cppflags = os.environ.get("CPPFLAGS", "")
            c_cpp_flags = shlex.split(env_cflags) + shlex.split(env_cppflags)
            if not any(opt.startswith("-g") for opt in c_cpp_flags):
                cflags += ["-g0"]
            # Linker options (Linux 64bit, gcc, tbb):
            # https://www.intel.com/content/www/us/en/developer/tools/oneapi/onemkl-link-line-advisor.html#gs.rhcqqp
            # -L${MKLROOT}/lib/intel64 -Wl, --no-as-needed -lmkl_intel_ilp64 -lmkl_tbb_thread
            # -lmkl_core -lpthread -lm -ldl
            # Compiler options:
            # -DMKL_ILP64 -m64 -I"${MKLROOT}/include"
            lib_dir = get_library_dir(check=True)
            mkl_libs = ["mkl_rt"]
            # Intel recommended link arg "--no-as-needed", but is not available on macOS,
            # and doesn't affect other OS, let's omit it.
            ldflags = _make_full_names(lib_dir, mkl_libs)
            libraries = ["nlopt", "pthread", "m", "dl"]

        self._add_cflags(cflags)
        self._add_ldflags(ldflags)
        self._add_libraries(libraries)

    def _add_cflags(self, flags: list[str]) -> None:
        self.extra_compile_args[:0] = flags

    def _add_ldflags(self, flags: list[str]) -> None:
        self.extra_link_args[:0] = flags

    def _add_libraries(self, libraries: list[str]) -> None:
        self.libraries[:0] = libraries



"""The mckit package code root."""
import mckit._init_dynamic_libraries as init_lib

from mckit._init_dynamic_libraries import (
    Body,
    Cone,
    Cylinder,
    GQuadratic,
    Plane,
    Shape,
    Sphere,
    Torus,
    create_surface,
)
from mckit.fmesh import FMesh
from mckit.material import AVOGADRO, Composition, Element, Material
from mckit.parser.mctal_parser import read_mctal
from mckit.parser.meshtal_parser import read_meshtal
from mckit.transformation import Transformation
from mckit.universe import Universe
from mckit.version import (
    __author__,
    __copyright__,
    __license__,
    __summary__,
    __title__,
    __version__,
)

__doc__ = __summary__



import os

import pytest


@pytest.fixture
def cd_tmpdir(tmpdir):
    """Temporarily switch to temp directory.

    Args:
        tmpdir: pytest fixture for temp directory

    Yields:
        None
    """
    old_dir = tmpdir.chdir()
    try:
        yield
    finally:
        os.chdir(old_dir)



from __future__ import annotations

import pytest

from mckit.card import Card


class DummyCard(Card):
    def mcnp_words(self, pretty=False):
        return [f"{k}: {v}" for k, v in self.options]


@pytest.mark.parametrize(
    "a, b, expected",
    [
        ({"a": 1}, {}, False),
        ({"a": 1}, {"a": 1}, True),
        ({"a": 1}, {"b": 1}, False),
        ({"a": 1}, {"a": 1, "b": 2}, False),
        ({"a": 1, "b": 2}, {"a": 1, "b": 2}, True),
        ({"a": 1, "b": 2}, {"a": 1}, False),
        ({"a": 1}, {"a": 2}, False),
    ],
)
def test_eq(a, b, expected):
    assert a == a
    assert b == b
    da, db = DummyCard(**a), DummyCard(**b)
    actual = da == db
    assert actual == expected
    assert (hash(da) == hash(db)) == expected



from typing import Iterable, Optional

from mckit.body import GLOBAL_BOX, Body, Card
from mckit.constants import MIN_BOX_VOLUME
from mckit.utils.indexes import Index, NumberedItemNotFoundError


class DummyCell(Body):
    """To substitute cell when it's not found."""

    def __init__(self, name: int):
        self._name = name
        options = {"name": name, "comment": "dummy"}
        Card.__init__(self, **options)

    def __hash__(self):
        return Card.__hash__(self)

    def __eq__(self, other):
        return Card.__eq__(self, other)

    def is_equivalent_to(self, other):
        return self.name() == other.name()

    @property
    def shape(self):
        raise NotImplementedError("The method is not available in dummy object")

    def intersection(self, other):
        raise NotImplementedError("The method is not available in dummy object")

    def union(self, other):
        raise NotImplementedError("The method is not available in dummy object")

    def simplify(
        self,
        box=GLOBAL_BOX,
        split_disjoint=False,
        min_volume=MIN_BOX_VOLUME,
        trim_size=1,
    ):
        raise NotImplementedError("The method is not available in dummy object")

    def fill(self, universe=None, recurrent=False, simplify=False, **kwargs):
        raise NotImplementedError("The method is not available in dummy object")

    def transform(self, tr):
        raise NotImplementedError("The method is not available in dummy object")


def raise_on_absent_cell_strategy(name: int) -> Optional[DummyCell]:
    raise CellNotFoundError(name)


def dummy_on_absent_cell_strategy(name: int) -> Optional[DummyCell]:
    return DummyCell(name)


class CellStrictIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(raise_on_absent_cell_strategy, **kwargs)

    @classmethod
    def from_iterable(cls, items: Iterable[Body]) -> "CellStrictIndex":
        index = cls()
        index.update((c.name(), c) for c in items)
        return index


class CellDummyIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(dummy_on_absent_cell_strategy, **kwargs)


class CellNotFoundError(NumberedItemNotFoundError):
    kind = "Cell"



# -*- coding: utf-8 -*-

import numpy as np

# noinspection PyUnresolvedReferences,PyPackageRequirements
from mckit.geometry import EX, EY, EZ, Box

from .transformation import Transformation
from .utils import mids


class AbstractMesh:
    pass


class RectMesh:
    """Represents rectangular mesh.

    Parameters
    ----------
    xbins, ybins, zbins : array_like[float]
        Bins of mesh in every direction. The last bin value gives dimension of
        mesh in this direction.
    transform : Trnasformation
        Transformation for the mesh. Default: None.

    Methods
    -------
    shape() - gets the shape of mesh.
    get_voxel(i, j, k) - gets the voxel of RectMesh with indices i, j, k.
    """

    def __init__(self, xbins, ybins, zbins, transform: Transformation = None):
        self._xbins = np.array(xbins)
        self._ybins = np.array(ybins)
        self._zbins = np.array(zbins)
        self._ex = EX
        self._ey = EY
        self._ez = EZ
        self._origin = np.array([self._xbins[0], self._ybins[0], self._zbins[0]])
        self._tr = None  # TODO dvp: and what this _tr is for?
        if transform is not None:
            self.transform(transform)  # TODO dvp: it's  wrong to apply transormaiton immideately

    def __eq__(self, other):
        return self is other

    def bounding_box(self):
        """Gets the bounding box of the cell.

        Returns
        -------
        bbox : Box
            Bounding box.
        """
        origin = [
            0.5 * (self._xbins[0] + self._xbins[-1]),
            0.5 * (self._ybins[0] + self._ybins[-1]),
            0.5 * (self._zbins[0] + self._zbins[-1]),
        ]
        if self._tr:
            origin = self._tr.apply2point(origin)
        dimx = self._xbins[-1] - self._xbins[0]
        dimy = self._ybins[-1] - self._ybins[0]
        dimz = self._zbins[-1] - self._zbins[0]
        return Box(origin, dimx, dimy, dimz, ex=self._ex, ey=self._ey, ez=self._ez)

    @property
    def shape(self):
        """Gets the shape of the mesh."""
        return self._xbins.size - 1, self._ybins.size - 1, self._zbins.size - 1

    def transform(self, tr: Transformation):
        """Transforms this mesh."""
        self._origin = tr.apply2point(self._origin)
        self._ex = tr.apply2vector(self._ex)
        self._ey = tr.apply2vector(self._ey)
        self._ez = tr.apply2vector(self._ez)
        if self._tr is not None:
            self._tr = tr.apply2transform(self._tr)
        else:
            self._tr = tr  # TODO dvp: this inconsistent with __init__, see TODO there

    def get_voxel(self, i, j, k):
        """Gets voxel.

        Parameters
        ----------
        i, j, k : int
            Indices of the voxel.

        Returns
        -------
        voxel : Box
            The box that describes the voxel.
        """
        cx = 0.5 * (self._xbins[i] + self._xbins[i + 1])
        cy = 0.5 * (self._ybins[j] + self._ybins[j + 1])
        cz = 0.5 * (self._zbins[k] + self._zbins[k + 1])
        center = np.array([cx, cy, cz])
        if self._tr:
            center = self._tr.apply2point(center)
        xdim = self._xbins[i + 1] - self._xbins[i]
        ydim = self._ybins[j + 1] - self._ybins[j]
        zdim = self._zbins[k + 1] - self._zbins[k]
        return Box(center, xdim, ydim, zdim, ex=self._ex, ey=self._ey, ez=self._ez)

    def voxel_index(self, point, local=False):
        """Gets index of voxel that contains specified point.

        Parameters
        ----------
        point : array_like[float]
            Coordinates of the point to be checked.
        local : bool
            If point is specified in local coordinate system.

        Returns
        -------
        i, j, k : int
            Indices along each dimension of voxel, where the point is located.
        """
        if self._tr and not local:
            point = self._tr.reverse().apply2point(point)
        else:
            point = np.array(point)
        x_proj = np.dot(point, EX)
        y_proj = np.dot(point, EY)
        z_proj = np.dot(point, EZ)
        i = np.searchsorted(self._xbins, x_proj) - 1
        j = np.searchsorted(self._ybins, y_proj) - 1
        k = np.searchsorted(self._zbins, z_proj) - 1
        if len(point.shape) == 1:
            return self.check_indices(i, j, k)
        else:
            print("i=", i, "j=", j, "k=", k)
            indices = []
            for x, y, z in zip(i, j, k):
                indices.append(self.check_indices(x, y, z))
            return indices

    def check_indices(self, i, j, k):
        """Check if the voxel with such indices really exists.

        Parameters
        ----------
        i, j, k : int
            Indices along x, y and z dimensions.

        Returns
        -------
        index_tuple : tuple(int)
            A tuple of indices if such voxel exists. None otherwise.
        """
        i = self._check_x(i)
        j = self._check_y(j)
        k = self._check_z(k)
        if i is None or j is None or k is None:
            return None
        else:
            return i, j, k

    def _check_x(self, i):
        if i < 0 or i >= self._xbins.size - 1:
            return None
        return i

    def _check_y(self, j):
        if j < 0 or j >= self._ybins.size - 1:
            return None
        return j

    def _check_z(self, k):
        if k < 0 or k >= self._zbins.size - 1:
            return None
        return k

    def slice_axis_index(self, X=None, Y=None, Z=None):
        """Gets index and axis of slice.

        Parameters
        ----------
        X, Y, Z : float
            Point of slice in local coordinate system.

        Returns
        -------
        axis : int
            Number of axis.
        index : int
            Index along axis.
        x, y : ndarray[float]
            Centers of bins along free axes.
        """
        none = 0
        for i, a in enumerate([X, Y, Z]):
            if a is not None:
                none += 1
                axis = i
        if none != 1:
            raise ValueError("Wrong number of fixed spatial variables.")

        if X is not None:
            index = self._check_x(np.searchsorted(self._xbins, X) - 1)
        elif Y is not None:
            index = self._check_y(np.searchsorted(self._ybins, Y) - 1)
        elif Z is not None:
            index = self._check_z(np.searchsorted(self._zbins, Z) - 1)
        else:
            index = None

        if index is None:
            raise ValueError("Specified point lies outside of the mesh.")

        if axis > 0:
            x = mids(self._xbins)
        else:
            x = mids(self._ybins)
        if axis < 2:
            y = mids(self._zbins)
        else:
            y = mids(self._ybins)

        return axis, index, x, y


class CylMesh:
    """Represents cylindrical mesh.

    Parameters
    ----------
    origin : array_like[float]
        Bottom center of the cylinder, that bounds the mesh.
    axis : array_like[float]
        Cylinder's axis.
    vec : array_like[float]
        Vector defining along with axis the plane for theta=0.
    rbins, zbins, tbins: array_like[float]
        Bins of mesh in radial, extend and angle directions respectively.
        Angles are specified in revolutions.
    """

    def __init__(self, origin, axis, vec, rbins, zbins, tbins):
        self._origin = np.array(origin)
        self._axis = np.array(axis)
        self._vec = np.array(vec)
        self._rbins = np.array(rbins)
        self._zbins = np.array(zbins)
        self._tbins = np.array(tbins)
        self._voxels = {}
        # raise NotImplementedError

    @property
    def shape(self):
        """Gets the shape of the mesh."""
        return self._rbins.size - 1, self._zbins.size - 1, self._tbins.size - 1

    def transform(self, tr: Transformation):
        """Transforms this mesh."""
        raise NotImplementedError

    def calculate_volumes(self, cells, with_mat_only=True, verbose=False, min_volume=1.0e-3):
        """Calculates volumes of cells.

        Parameters
        ----------
        cells : list[Body]
            List of cells.
        verbose : bool
            Verbose output during calculations.
        min_volume : float
            Minimum volume for cell volume calculations

        Returns
        -------
        volumes : dict
            Volumes of cells for every voxel. It is dictionary cell -> vol_matrix.
            vol_matrix is SparseData instance - volume of cell for each voxel.
        """
        raise NotImplementedError

    def get_voxel(self, i, j, k):
        """Gets voxel.

        Parameters
        ----------
        i, j, k : int
            Indices of the voxel.

        Returns
        -------
        voxel : Box
            The box that describes the voxel.
        """
        raise NotImplementedError

    def voxel_index(self, point, local=False):
        """Gets index of voxel that contains specified point.

        Parameters
        ----------
        point : array_like[float]
            Coordinates of the point to be checked.
        local : bool
            If point is specified in local coordinate system.

        Returns
        -------
        i, j, k : int
            Indices along each dimension of voxel, where the point is located.
        """
        raise NotImplementedError

    def check_indices(self, i, j, k):
        """Check if the voxel with such indices really exists.

        Parameters
        ----------
        i, j, k : int
            Indices along x, y and z dimensions.

        Returns
        -------
        index_tuple : tuple(int)
            A tuple of indices if such voxel exists. None otherwise.
        """
        i = self._check_r(i)
        j = self._check_z(j)
        k = self._check_t(k)
        if i is None or j is None or k is None:
            return None
        else:
            return i, j, k

    def _check_r(self, i):
        if i < 0 or i >= self._rbins.size - 1:
            return None
        return i

    def _check_z(self, j):
        if j < 0 or j >= self._zbins.size - 1:
            return None
        return j

    def _check_t(self, k):
        if k < 0 or k >= self._tbins.size - 1:
            return None
        return k

    def slice_axis_index(self, R=None, Z=None, T=None):
        """Gets index and axis of slice.

        Parameters
        ----------
        R, Z, T : float
            Point of slice in local coordinate system.

        Returns
        -------
        axis : int
            Number of axis.
        index : int
            Index along axis.
        x, y : ndarray[float]
            Centers of bins along free axes.
        """
        raise NotImplementedError


class FMesh:
    """Fmesh tally object.

    Parameters
    ----------
    name : int
        Tally name.
    particle : str
        Particle type (neutron, photon, electron).
    ebins : array_like[float]
        Bin boundaries for energies.
    xbins, ybins, zbins : array_like[float]
        Bin boundaries in X, Y and Z directions respectively for rectangular mesh.
    rbins, zbins, tbins : array_like[float]
        Bin boundaries in R, Z and Theta directions for cylindrical mesh.
    origin : array_like[float]
        Bottom center of the cylinder (For cylindrical mesh only).
    axis : array_like[float]
        Axis of the cylinder (for cylindrical mesh only).
    vec : array_like[float]
        Vector defining along with axis the plane for theta=0.
    data : arraylike[float]
        Fmesh data - quantity estimation averaged over voxel volume. It has
        shape (Ne-1)x(Nx-1)x(Ny-1)x(Nz-1), where Ne, Nx, Ny and Nx - the number
        of corresponding bin boundaries.
    error : arraylike[float]
        Fmesh relative error. Shape - see data.
    transform : Transformation
        Transformation to be applied to the spatial mesh.
    histories : int
        The number of histories run to obtain meshtal data.
    modifier : None
        Data transformation.

    Methods
    -------
    get_slice()
        Gets specific slice of data.
    get_spectrum(point)
        Gets energy spectrum at the specified point.
    get_spectrum_by_index(index)
        Gets energy spectrum at the specified mesh index.
    mean_flux()
        Gets average flux for every energy bin.
    """

    def __init__(
        self,
        name,
        particle,
        data,
        error,
        ebins=None,
        xbins=None,
        ybins=None,
        zbins=None,
        rbins=None,
        tbins=None,
        dtbins=None,
        transform: Transformation = None,
        modifier=None,
        origin=None,
        axis=None,
        vec=None,
        histories=None,
    ):
        self._data = np.array(data)
        self._error = np.array(error)
        self._name = name
        self._histories = histories
        self._particle = particle
        if ebins is not None:
            self._ebins = np.array(ebins)
        else:
            self._ebins = np.array([0, 1.0e36])
        if dtbins is not None:
            self._dtbins = np.array(dtbins)
        self._modifier = modifier
        if rbins is None and tbins is None:
            self._mesh = RectMesh(xbins, ybins, zbins, transform=transform)
        elif xbins is None and ybins is None:
            self._mesh = CylMesh(origin, axis, vec, rbins, zbins, tbins)
        if self._data.shape[1:] != self._mesh.shape:
            raise ValueError("Incorrect data shape")
        elif self._error.shape[1:] != self._mesh.shape:
            raise ValueError("Incorrect error shape")

    @property
    def mesh(self):
        return self._mesh

    @property
    def particle(self):
        return self._particle

    @property
    def histories(self):
        """The number of histories in the run."""
        return self._histories

    def mean_flux(self):
        """Gets average flux.

        Returns
        -------
        ebins : np.array[float]
            Energy bin boundaries.
        flux : np.array[float]
            Average flux in each energy bin.
        """
        return self._ebins.copy(), np.mean(self._data, axis=(1, 2, 3))

    def get_spectrum(self, point):
        """Gets energy spectrum at the specified point.

        Parameters
        ----------
        point : arraylike[float]
            Point energy spectrum must be get at.

        Returns
        -------
        energies: ndarray[float]
            Energy bins for the spectrum at the point - group boundaries.
        flux : ndarray[float]
            Group flux at the point.
        err : ndarray[float]
            Relative errors for flux components.
        """
        index = self._mesh.voxel_index(point)
        if index is None:
            raise ValueError("Point {0} lies outside of the mesh.".format(point))
        return self.get_spectrum_by_index(index)

    def get_spectrum_by_index(self, index):
        """Gets energy spectrum in the specified voxel.

        Parameters
        ----------
        index : tuple[int]
            Indices of spatial mesh bins.

        Returns
        -------
        energies: ndarray[float]
            Energy bins for the spectrum at the point - group boundaries.
        flux : ndarray[float]
            Group flux at the point.
        err : ndarray[float]
            Relative errors for flux components.
        """
        i, j, k = index
        flux = self._data[:, i, j, k]
        err = self._error[:, i, j, k]
        return self._ebins, flux, err

    def get_slice(self, E="total", X=None, Y=None, Z=None, R=None, T=None):
        """Gets data in the specified slice. Only one spatial letter is allowed.

        Parameters
        ----------
        E : str or float
            Energy value of interest. It specifies energy bin. If 'total' - then data
            is summed across energy axis.
        X, Y, Z : float
            Spatial point which belongs to the slice plane. Other two dimensions are free.

        Returns
        -------
        x, y : ndarray[float]
            Centers of spatial bins in free directions.
        data : ndarray[float]
            Data
        err : ndarray[float]
            Relative errors for data.
        """
        if isinstance(self._mesh, RectMesh):
            axis, index, x, y = self._mesh.slice_axis_index(X=X, Y=Y, Z=Z)
        else:
            axis, index, x, y = self._mesh.slice_axis_index(R=R, Z=Z, T=T)

        data = self._data.take(index, axis=axis + 1)  # +1 because the first axis is for energy.
        err = self._error.take(index, axis=axis + 1)

        if E == "total":
            abs_err = (data * err) ** 2
            abs_tot_err = np.sqrt(np.sum(abs_err, axis=0))
            data = np.sum(data, axis=0)
            err = np.nan_to_num(abs_tot_err / data)
        else:
            if E <= self._ebins[0] or E > self._ebins[-1]:
                raise ValueError("Specified energy lies outside of energy bins.")
            i = np.searchsorted(self._ebins, E) - 1
            data = data.take(i, axis=0)
            err = err.take(i, axis=0)
        return x, y, data, err



import typing as tp

from io import StringIO
from pathlib import Path

import click

from mckit.cli.logging import logger
from mckit.constants import MCNP_ENCODING
from mckit.universe import Universe, UniverseAnalyser

# This is the encoding swallowing non ascii (neither unicode) symbols happening in MCNP models code


def check_if_path_exists(path: tp.Union[str, Path], override: bool):
    if isinstance(path, str):
        path = Path(path)
    if not override and path.exists():
        errmsg = f"""\
        Cannot override existing file \"{path}\".
        Please remove the file or specify --override option"""
        logger.error(errmsg)
        raise click.UsageError(errmsg)


def save_mcnp(model: Universe, path: tp.Union[str, Path], override: bool):
    check_if_path_exists(path, override)
    analyser = UniverseAnalyser(model)
    if analyser.we_are_all_clear():
        model.save(path, encoding=MCNP_ENCODING, check_clashes=False)
    else:
        out = StringIO()
        out.write("Duplicates found:\n")
        analyser.print_duplicates_map(stream=out)
        raise ValueError(out.getvalue())


def get_default_output_directory(source, suffix):
    return Path(Path(source).with_suffix(suffix).name)



from typing import NamedTuple

import mckit.utils.named as nm
import pytest

from mckit.utils.indexes import (
    IndexOfNamed,
    NumberedItemDuplicateError,
    StatisticsCollector,
    ignore_equal_objects_strategy,
    raise_on_duplicate_strategy,
)


class Something(NamedTuple):
    name: int


@pytest.mark.parametrize(
    "entities, expected",
    [
        ([], {}),
        (
            [Something(i) for i in range(1, 3)],
            {k: Something(k) for k in range(1, 3)},
        ),
    ],
)
def test_index_of_named_happy_path(entities, expected):
    actual = IndexOfNamed.from_iterable(entities, key=lambda x: x.name)
    assert actual == expected


@pytest.mark.parametrize(
    "entities",
    [
        (list(map(Something, [1, 1]))),
    ],
)
def test_clashes(entities):
    with pytest.raises(NumberedItemDuplicateError, match="Something"):
        actual = IndexOfNamed.from_iterable(
            entities, key=lambda x: x.name, on_duplicate=raise_on_duplicate_strategy
        )
        assert not actual


@pytest.mark.parametrize(
    "entities, expected",
    [
        (list(map(Something, [1, 1])), {1: Something(1)}),
    ],
)
def test_equal_duplicates(entities, expected):
    actual = IndexOfNamed.from_iterable(
        entities, key=lambda x: x.name, on_duplicate=ignore_equal_objects_strategy
    )
    assert actual == expected


class Something2(NamedTuple):
    name: nm.Name
    value: int


@pytest.mark.parametrize(
    "entities",
    [
        [Something2(1, 1), Something2(1, 2)],
    ],
)
def test_clashes_on_non_equal_items(entities):
    with pytest.raises(NumberedItemDuplicateError, match="Something2"):
        actual = IndexOfNamed.from_iterable(
            entities, key=lambda x: x.name, on_duplicate=raise_on_duplicate_strategy
        )
        assert not actual


@pytest.mark.parametrize(
    "entities, expected, expected_collected",
    [
        ([Something2(1, 1), Something2(1, 2)], {1: Something2(1, 2)}, {1: 2}),
        (
            [Something2(1, 1), Something2(1, 2), Something2(2, 3)],
            {1: Something2(1, 2), 2: Something2(2, 3)},
            {1: 2},
        ),
        (
            [Something2(1, 1), Something2(1, 2), Something2(2, 3), Something2(2, 4)],
            {1: Something2(1, 2), 2: Something2(2, 4)},
            {1: 2, 2: 2},
        ),
    ],
)
def test_collect_statistics_on_clashes(entities, expected, expected_collected):
    collector = StatisticsCollector()
    actual = IndexOfNamed.from_iterable(entities, key=lambda x: x.name, on_duplicate=collector)
    assert actual == expected
    assert collector == expected_collected



# -*- coding: utf-8 -*-
"""Сборка модели из конвертов и входяших в них юниверсов по заданной спецификации."""
from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

from functools import reduce
from pathlib import Path

import mckit as mk
import numpy as np

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from mckit import Transformation
from mckit.cli.logging import logger
from mckit.parser.mcnp_input_sly_parser import ParseResult, from_file
from mckit.utils import filter_dict

from .common import save_mcnp


def compose(output, fill_descriptor_path, source, override):
    logger.info("Loading model from {s}", s=source)
    parse_result: ParseResult = from_file(source)
    envelopes = parse_result.universe
    source = Path(source)
    universes_dir = source.absolute().parent
    assert universes_dir.is_dir()
    logger.info("Loading fill-descriptor from {f}", f=fill_descriptor_path)
    with fill_descriptor_path.open("rb") as fid:
        fill_descriptor = tomllib.load(fid)

    universes = load_universes(fill_descriptor, universes_dir)
    named_transformations = load_named_transformations(fill_descriptor)

    comps = {}

    for k, v in universes.items():
        u, _ = v
        cps = u.get_compositions()
        comps[k] = {c for c in cps}

    common = reduce(set.union, comps.values())
    envelopes.set_common_materials(common)
    cells_index = dict((cell.name(), cell) for cell in envelopes)

    for i, spec in universes.items():
        universe, transformation = spec
        universe.set_common_materials(common)
        cell = cells_index[i]
        cell.options = filter_dict(cell.options, "original")
        cell.options["FILL"] = {"universe": universe}
        if transformation is not None:
            # if isinstance(transformation, tk_items.Array):
            if isinstance(transformation, list):
                transformation1 = np.fromiter(map(float, iter(transformation)), dtype=np.double)
                try:
                    translation = transformation1[:3]
                    if len(transformation1) > 3:
                        rotation = transformation1[3:]
                    else:
                        rotation = None
                    transformation2 = mk.Transformation(
                        translation=translation,
                        rotation=rotation,
                        indegrees=True,  # Assuming that on decompose we store a transformation in degrees as well
                    )
                except ValueError as ex:
                    raise ValueError(
                        f"Failed to process FILL transformation in cell #{cell.name()} of universe #{universe.name()}"
                    ) from ex
                cell.options["FILL"]["transform"] = transformation2
            # elif isinstance(transformation, tk_items.Integer):
            elif isinstance(transformation, int):
                assert (
                    named_transformations is not None
                ), "There are no named transformations in the fill descriptor file"
                transformation1 = named_transformations[int(transformation)]
                cell.options["FILL"]["transform"] = transformation1
            else:
                raise NotImplementedError(
                    f"Unexpected type of transformation parameter {type(transformation)}"
                )
    save_mcnp(envelopes, output, override)


def load_universes(
    fill_descriptor, universes_dir
) -> Dict[int, Tuple[mk.Universe, Union[int, List[float]]]]:
    filler_path_map: Dict[int, Tuple[Path, mk.Universe]] = dict()
    cell_filler_map: Dict[int, Tuple[mk.Universe, Union[int, List[float]]]] = dict()

    for k, v in fill_descriptor.items():
        if isinstance(v, dict) and "universe" in v:
            cell_name = int(k)
            universe_name = int(v["universe"])
            transformation = v.get("transform", None)
            universe_path = Path(v["file"])
            if universe_name in filler_path_map:
                prev_path, prev_universe = filler_path_map[universe_name]
                if prev_path == universe_path:
                    universe = prev_universe
                else:
                    raise ValueError(
                        f'\n \
                        Filler number {universe_name} specifies different paths to load \n \
                        "{prev_path}" and "{universe_path}"'
                    )
            else:
                if universe_path.exists():
                    load_path = universe_path
                else:
                    load_path = universes_dir / universe_path
                if not load_path.exists():
                    raise FileNotFoundError(universe_path)
                logger.info("Loading file {u}", u=load_path)
                parse_result: ParseResult = from_file(load_path)
                universe = parse_result.universe
                universe.rename(name=universe_name)
                filler_path_map[universe_name] = (universe_path, universe)

            cell_filler_map[cell_name] = (universe, transformation)

    return cell_filler_map


def load_named_transformations(fill_descriptor) -> Optional[Dict[int, Transformation]]:
    transformations = fill_descriptor.get("named_transformations", None)
    if transformations:
        named_transformations = {}
        for k, v in transformations.items():
            name = int(k[2:])
            transform_params = np.fromiter(map(float, v), dtype=float)
            translation = transform_params[:3]
            if transform_params.size == 9:
                rotation = transform_params[3:]
            else:
                rotation = None
            transform = Transformation(
                translation=translation, rotation=rotation, indegrees=True, name=name
            )
            named_transformations[name] = transform
        return named_transformations
    else:
        return None



from __future__ import annotations

from typing import Generator

from pathlib import Path

MCNP_ENCODING = "Cp1251"
"""The encoding used in SuperMC when creating MCNP models code.

Some symbols are not Unicode.
"""


def make_dir(d: Path) -> Path:
    """Create directory."""
    d.mkdir(parents=True, exist_ok=True)
    return d


def make_dirs(*dirs: Path) -> Generator[Path, None, None]:
    yield from (make_dir(f) for f in dirs)


def check_if_path_exists(p: Path) -> Path:
    if p.exists():
        return p
    raise FileNotFoundError(f'Path "{p}" does not exist')


def check_if_all_paths_exist(*paths: Path) -> Generator[Path, None, None]:
    yield from map(check_if_path_exists, paths)



"""The interfaces to facilitate adding new functionality to an hierarchy of classes (Visitor pattern)."""
from typing import Any, Callable, ContextManager, NoReturn

from functools import reduce

TVisitor = Callable[..., ContextManager]


def accept(acceptor: Any, visitor: TVisitor, *args, **kwargs) -> Any:
    with visitor(acceptor, *args, **kwargs) as recurse:
        items_visitor, initial = recurse
        return reduce(items_visitor, acceptor, initial)


def on_unknown_acceptor(acceptor: Any) -> NoReturn:
    raise NotImplementedError(f"Not implemented for {type(acceptor).__name__}")



from io import StringIO

import pytest

from mckit.cli.logging import logger
from mckit.cli.runner import VERSION, mckit, meta


def test_version_command(runner):
    result = runner.invoke(mckit, args=["--version"], catch_exceptions=False)
    assert result.exit_code == 0, "Should success on '--version' option: " + result.output
    assert VERSION in result.output, "print version on 'version' command"


def test_help_command(runner):
    result = runner.invoke(mckit, args=["--help"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "Usage: " in result.output
    assert meta.__summary__ in result.output


def test_fails_on_unknown_command(runner):
    result = runner.invoke(mckit, args=["unknown_command"], catch_exceptions=False)
    assert result.exit_code != 0, "Have somebody implemented 'unknown_command' already?"
    assert "Usage: " in result.output


def test_writes_to_logger_on_errors(runner):
    sink = StringIO()
    logger.add(sink)
    result = runner.invoke(mckit, args=["check", "not_existing.i"], catch_exceptions=False)
    assert result.exit_code != 0, "Does not_existing.i exist?"


if __name__ == "__main__":
    pytest.main()



"""Surface methods."""
from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Text, Tuple, Union

from abc import abstractmethod

import numpy as np

from mckit.box import GLOBAL_BOX

# fmt:off
# noinspection PyUnresolvedReferences,PyPackageRequirements
from mckit.geometry import BOX as _BOX
from mckit.geometry import EX, EY, EZ, ORIGIN
from mckit.geometry import RCC as _RCC
from mckit.geometry import Cone as _Cone
from mckit.geometry import Cylinder as _Cylinder
from mckit.geometry import GQuadratic as _GQuadratic
from mckit.geometry import Plane as _Plane
from mckit.geometry import Sphere as _Sphere
from mckit.geometry import Torus as _Torus

from . import constants
from .card import Card
from .constants import DROP_OPTIONS
from .printer import add_float, pretty_float
from .transformation import Transformation
from .utils import (
    are_equal,
    compute_hash,
    deepcopy,
    filter_dict,
    round_array,
    round_scalar,
    significant_array,
    significant_digits,
)
from .utils.tolerance import DEFAULT_TOLERANCE_ESTIMATOR, FLOAT_TOLERANCE, MaybeClose

# fmt:on


# noinspection PyUnresolvedReferences,PyPackageRequirements
__all__ = [
    "create_surface",
    "Plane",
    "Sphere",
    "Cone",
    "Torus",
    "GQuadratic",
    "Cylinder",
    "Surface",
    "RCC",
    "BOX",
    "ORIGIN",
    "EX",
    "EY",
    "EZ",
]


VectorLike = Union[np.ndarray, List[float]]


# noinspection PyPep8Naming
def create_surface(kind: str, *params: float, **options: Dict[str, Any]) -> "Surface":
    """Creates new surface.

    Args:
        kind: Surface kind designator. See MCNP manual.
        params: List of surface parameters.
        options: Dictionary of surface's options.
                In particular, transform  - transformation instance to be applied to the surface being created.

    Returns:
        New surface.

    Raises:
        NotImplementedError: when some logic is not implemented yet
        ValueError: on incompatible `params`
    """
    params = np.asarray(params, dtype=float)
    kind = kind.upper()
    if kind[-1] == "X":
        axis = EX
        assume_normalized = True
    elif kind[-1] == "Y":
        axis = EY
        assume_normalized = True
    elif kind[-1] == "Z":
        axis = EZ
        assume_normalized = True
    else:
        axis = None
        assume_normalized = False
    # -------- Plane -------------------
    if kind[0] == "P":
        if len(kind) == 2:
            return Plane(axis, -params[0], assume_normalized=assume_normalized, **options)
        else:
            return Plane(
                params[:3],
                -params[3],
                assume_normalized=assume_normalized,
                **options,
            )
    # -------- SQ -------------------
    elif kind == "SQ":
        A, B, C, D, E, F, G, x0, y0, z0 = params
        m = np.diag([A, B, C])
        v = 2 * np.array([D - A * x0, E - B * y0, F - C * z0])
        k = A * x0**2 + B * y0**2 + C * z0**2 - 2 * (D * x0 + E * y0 + F * z0) + G
        return GQuadratic(m, v, k, **options)
    # -------- Sphere ------------------
    elif kind[0] == "S":
        if kind == "S":
            r0 = np.array(params[:3])
        elif kind == "SO":
            r0 = ORIGIN
        else:
            r0 = axis * params[0]
        R = params[-1]
        return Sphere(r0, R, **options)
    # -------- Cylinder ----------------
    elif kind[0] == "C":
        A = 1 - axis
        if kind[1] == "/":
            Ax, Az = np.dot(A, EX), np.dot(A, EZ)
            r0 = params[0] * (Ax * EX + (1 - Ax) * EY) + params[1] * ((1 - Az) * EY + Az * EZ)
        else:
            r0 = ORIGIN
        R = params[-1]
        return Cylinder(r0, axis, R, assume_normalized=assume_normalized, **options)
    # -------- Cone ---------------
    elif kind[0] == "K":
        if kind[1] == "/":
            r0 = np.array(params[:3], dtype=float)
            ta = params[3]
        else:
            r0 = params[0] * axis
            ta = params[1]
        sheet = 0 if len(params) % 2 == 0 else int(params[-1])
        return Cone(r0, axis, ta, sheet, assume_normalized=assume_normalized, **options)
    # ---------- GQ -----------------
    elif kind == "GQ":
        A, B, C, D, E, F, G, H, J, k = params
        m = np.array([[A, 0.5 * D, 0.5 * F], [0.5 * D, B, 0.5 * E], [0.5 * F, 0.5 * E, C]])
        v = np.array([G, H, J])
        return GQuadratic(m, v, k, **options)
    # ---------- Torus ---------------------
    elif kind[0] == "T":
        x0, y0, z0, R, a, b = params
        return Torus([x0, y0, z0], axis, R, a, b, **options)
    # ---------- Macrobodies ---------------
    elif kind == "RPP":
        x_min, x_max, y_min, y_max, z_min, z_max = params
        center = [x_min, y_min, z_min]
        dir_x = [x_max - x_min, 0, 0]
        dir_y = [0, y_max - y_min, 0]
        dir_z = [0, 0, z_max - z_min]
        return BOX(center, dir_x, dir_y, dir_z, **options)
    elif kind == "BOX":
        center = params[:3]
        dir_x = params[3:6]
        dir_y = params[6:9]
        dir_z = params[9:]
        return BOX(center, dir_x, dir_y, dir_z, **options)
    elif kind == "RCC":
        center = params[:3]
        axis = params[3:6]
        radius = params[6]
        return RCC(center, axis, radius, **options)
    # ---------- Axis-symmetric surface defined by points ------
    else:
        if len(params) == 2:
            return Plane(axis, -params[0], **options)
        elif len(params) == 4:
            # TODO: Use special classes instead of GQ
            h1, r1, h2, r2 = params
            if abs(h2 - h1) < constants.RESOLUTION * max(abs(h1), abs(h2)):
                return Plane(axis, -0.5 * (h1 + h2), **options)
            elif abs(r2 - r1) < constants.RESOLUTION * max(abs(r2), abs(r1)):
                R = 0.5 * (abs(r1) + abs(r2))
                return Cylinder([0, 0, 0], axis, R, **options)
            else:
                if r1 * r2 < 0:
                    raise ValueError("Points must belong to the one sheet.")
                h0 = (abs(r1) * h2 - abs(r2) * h1) / (abs(r1) - abs(r2))
                t2 = ((r1 - r2) / (h1 - h2)) ** 2
                s = int(
                    round((h1 - h0) / abs(h1 - h0))
                )  # TODO: dvp check this conversion: was without int()
                return Cone(axis * h0, axis, t2, sheet=s, **options)
        elif len(params) == 6:
            # TODO: Implement creation of surface by 3 points.
            raise NotImplementedError


def create_replace_dictionary(
    surfaces: Set["Surface"],
    unique: Optional[set["Surface"]] = None,
    box=GLOBAL_BOX,
    tol: float = 1.0e-10,
) -> dict["Surface", tuple["Surface", int]]:
    """Creates surface replace dictionary for equal surfaces removing.

    Args:
        surfaces : A set of surfaces to be checked.
        unique:  A set of surfaces that are assumed to be unique. If not None, then
                `surfaces` are checked for coincidence with one of them.
        box : Box
            A box, which is used for comparison.
        tol : float
            Tolerance

    Returns:
        A replacement dictionary. surface -> (replace_surface, sense). Sense is +1
        if surfaces have the same direction of normals. -1 otherwise.
    """
    replace = {}
    unique_surfaces = set() if unique is None else unique
    for s in surfaces:
        for us in unique_surfaces:
            t = s.equals(us, box=box, tol=tol)
            if t != 0:
                replace[s] = (us, t)
                break
        else:
            unique_surfaces.add(s)
    return replace


def _drop_empty_transformation(options: Dict[str, Any]) -> None:
    if "transform" in options and not options["transform"]:  # empty transformation option
        del options["transform"]


class Surface(Card, MaybeClose):
    """Base class for all surface classes.

    Methods
    -------
    equals(other, box, tol)
        Checks if this surface and surf are equal inside the box.
    test_point(p)
        Checks the sense of point p with respect to this surface.
    test_box(box)
        Checks whether this surface crosses the box.
    projection(p)
        Gets projection of point p on the surface.
    """

    def __init__(self, **options) -> None:
        """Create :class:`Surface` with the options.

        Args:
            options: kwargs - properties for the `Surface` card.
        """
        _drop_empty_transformation(options)
        Card.__init__(self, **options)

    @abstractmethod
    def copy(self) -> "Surface":
        pass

    @property
    def transformation(self) -> Optional[Transformation]:
        return self.options.get("transform", None)

    @abstractmethod
    def apply_transformation(self) -> "Surface":
        """Applies transformation specified for the surface.

        Returns
        -------
        A 1new surface with transformed parameters, If there's specified transformation,
        otherwise returns self.
        """

    def combine_transformations(self, tr: Transformation) -> Transformation:
        my_transformation: Transformation = self.transformation
        if my_transformation:
            tr = tr.apply2transform(my_transformation)
        return tr

    @abstractmethod
    def transform(self, tr: Transformation) -> "Surface":
        """Applies transformation to this surface.

        Args:
            tr: Transformation to be applied.

        Returns:
            The result of this surface transformation.
        """

    @abstractmethod
    def is_close_to(
        self,
        other: "Surface",
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        """Checks if this surface is close to other one with the given tolerance values."""

    @abstractmethod
    def round(self) -> "Surface":
        """Returns rounded version of self."""

    def mcnp_words(self, pretty: bool = False) -> list[str]:
        words = []
        mod = self.options.get("modifier", None)
        if mod:
            words.append(mod)
        words.append(str(self.name()))
        words.append(" ")
        # TODO dvp: add transformations processing in Universe.
        return words

    def clean_options(self) -> Dict[Text, Any]:
        result: Dict[Text, Any] = filter_dict(self.options, DROP_OPTIONS)
        return result

    def compare_transformations(self, tr: Transformation) -> bool:
        my_transformation = self.transformation
        if my_transformation is not None:
            if tr is None:
                return False
            return my_transformation == tr
        else:
            return tr is None

    def has_close_transformations(
        self, tr: Transformation, estimator: Callable[[Any, Any], bool]
    ) -> bool:
        my_transformation = self.transformation
        if my_transformation is None:
            return tr is None
        if tr is None:
            return False
        return my_transformation.is_close(tr, estimator)

    def __getstate__(self):
        return self.options

    def __setstate__(self, state):
        self.options = state


def internalize_ort(v: np.ndarray) -> Tuple[np.ndarray, bool]:
    if v is EX or np.array_equal(v, EX):
        return EX, True
    elif v is EY or np.array_equal(v, EY):
        return EY, True
    elif v is EZ or np.array_equal(v, EZ):
        return EZ, True
    return v, False


class RCC(Surface, _RCC):
    def __init__(self, center, direction, radius, **options):
        center = np.array(center)
        direction = np.array(direction)
        opt_surf = options.copy()
        opt_surf["name"] = 1
        norm = np.array(direction) / np.linalg.norm(direction)
        cyl = Cylinder(center, norm, radius, **opt_surf).apply_transformation()
        center2 = center + direction
        offset2 = -np.dot(norm, center2).item()
        offset3 = np.dot(norm, center).item()
        opt_surf["name"] = 2
        plane2 = Plane(norm, offset2, **opt_surf).apply_transformation()
        opt_surf["name"] = 3
        plane3 = Plane(-norm, offset3, **opt_surf).apply_transformation()
        _RCC.__init__(self, cyl, plane2, plane3)
        options.pop("transform", None)
        Surface.__init__(self, **options)
        self._hash = hash(cyl) ^ hash(plane2) ^ hash(plane3)

    def surface(self, number):
        args = self.surfaces
        if 1 <= number <= len(args):
            return args[number - 1]
        else:
            raise ValueError(f"There is no such surface in macrobody: {number}")

    def get_params(self):
        args = self.surfaces
        for a in args:
            print(a.mcnp_repr())
        center = args[0]._pt - args[2]._k * args[0]._axis * np.dot(args[0]._axis, args[2]._v)
        direction = -(args[1]._k + args[2]._k) * args[1]._v
        radius = args[0]._radius
        return center, direction, radius

    def mcnp_words(self, pretty: bool = False) -> list[str]:
        words = Surface.mcnp_words(self, pretty)
        words.append("RCC")
        words.append(" ")
        center, direction, radius = self.get_params()
        values = list(center)
        values.extend(direction)
        values.append(radius)
        for v in values:
            fd = significant_digits(
                v, constants.FLOAT_TOLERANCE, resolution=constants.FLOAT_TOLERANCE
            )
            words.append(pretty_float(v, fd))
            words.append(" ")
        return words

    def transform(self, tr: Transformation) -> "RCC":
        """Transforms the shape.

        Args:
            tr:  Transformation to be applied.

        Returns:
            New RCC shape with the transformation stored in options.
        """
        center, direction, radius = self.get_params()
        # TODO(dvp): What if `self` already has transformation?
        #            Should we apply the both transformation on new one?
        return RCC(center, direction, radius, transform=tr)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, RCC):
            return False
        args_this = self.surfaces
        args_other = other.surfaces
        return args_this == args_other

    def __getstate__(self):
        surf_state = Surface.__getstate__(self)
        args = self.surfaces
        return args, self._hash, surf_state

    def __setstate__(self, state):
        args, hash_value, surf_state = state
        Surface.__setstate__(self, surf_state)
        _RCC.__init__(self, *args)
        self._hash = hash_value

    def copy(self):
        center, direction, radius = self.get_params()
        return RCC(center, direction, radius, **self.options)


class BOX(Surface, _BOX):
    """Macrobody BOX surface.

    Parameters
    ----------
    """

    def __init__(self, center, dir_x, dir_y, dir_z, **options):
        dir_x = np.array(dir_x)
        dir_y = np.array(dir_y)
        dir_z = np.array(dir_z)
        center = np.array(center)
        center2 = center + dir_x + dir_y + dir_z
        len_x = np.linalg.norm(dir_x)
        len_y = np.linalg.norm(dir_y)
        len_z = np.linalg.norm(dir_z)
        norm_x = dir_x / len_x
        norm_y = dir_y / len_y
        norm_z = dir_z / len_z
        offset_x = np.dot(norm_x, center).item()
        offset_y = np.dot(norm_y, center).item()
        offset_z = np.dot(norm_z, center).item()
        offset_2x = -np.dot(norm_x, center2).item()
        offset_2y = -np.dot(norm_y, center2).item()
        offset_2z = -np.dot(norm_z, center2).item()
        opt_surf = options.copy()
        opt_surf["name"] = 1
        surf1 = Plane(norm_x, offset_2x, **opt_surf).apply_transformation()
        opt_surf["name"] = 2
        surf2 = Plane(-norm_x, offset_x, **opt_surf).apply_transformation()
        opt_surf["name"] = 3
        surf3 = Plane(norm_y, offset_2y, **opt_surf).apply_transformation()
        opt_surf["name"] = 4
        surf4 = Plane(-norm_y, offset_y, **opt_surf).apply_transformation()
        opt_surf["name"] = 5
        surf5 = Plane(norm_z, offset_2z, **opt_surf).apply_transformation()
        opt_surf["name"] = 6
        surf6 = Plane(-norm_z, offset_z, **opt_surf).apply_transformation()
        _BOX.__init__(self, surf1, surf2, surf3, surf4, surf5, surf6)
        options.pop("transform", None)
        Surface.__init__(self, **options)
        self._hash = (
            hash(surf1) ^ hash(surf2) ^ hash(surf3) ^ hash(surf4) ^ hash(surf5) ^ hash(surf6)
        )

    def surface(self, number):
        args = self.surfaces
        if 1 <= number <= len(args):
            return args[number - 1]
        else:
            raise ValueError(f"There is no such surface in macrobody: {number}")

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, BOX):
            return False
        args_this = self.surfaces
        args_other = other.surfaces
        return args_this == args_other

    @staticmethod
    def _get_plane_intersection(s1, s2, s3):
        matrix = np.zeros((3, 3))
        matrix[0, :] = -s1._v
        matrix[1, :] = -s2._v
        matrix[2, :] = -s3._v
        vector = np.array([s1._k, s2._k, s3._k])
        return np.linalg.solve(matrix, vector)

    def get_params(self):
        args = self.surfaces
        center = self._get_plane_intersection(args[1], args[3], args[5])
        point2 = self._get_plane_intersection(args[0], args[2], args[4])
        norm_x = args[0]._v
        norm_y = args[2]._v
        norm_z = args[4]._v
        diag = point2 - center
        dir_x = np.dot(norm_x, diag) * norm_x
        dir_y = np.dot(norm_y, diag) * norm_y
        dir_z = np.dot(norm_z, diag) * norm_z
        return center, dir_x, dir_y, dir_z

    def mcnp_words(self, pretty: bool = False) -> list[str]:
        words = Surface.mcnp_words(self, pretty)
        words.append("BOX")
        words.append(" ")
        center, dir_x, dir_y, dir_z = self.get_params()
        values = list(center)
        values.extend(dir_x)
        values.extend(dir_y)
        values.extend(dir_z)
        for v in values:
            fd = significant_digits(
                v, constants.FLOAT_TOLERANCE, resolution=constants.FLOAT_TOLERANCE
            )
            words.append(pretty_float(v, fd))
            words.append(" ")
        return words

    def transform(self, tr):
        """Transforms the shape.

        Parameters
        ----------
        tr : Transformation
            Transformation to be applied.

        Returns
        -------
        result : Shape
            New shape.
        """
        center, dir_x, dir_y, dir_z = self.get_params()
        return BOX(center, dir_x, dir_y, dir_z, transform=tr)

    def __getstate__(self):
        surf_state = Surface.__getstate__(self)
        args = self.surfaces
        return args, self._hash, surf_state

    def __setstate__(self, state):
        args, hash_value, surf_state = state
        Surface.__setstate__(self, surf_state)
        _BOX.__init__(self, *args)
        self._hash = hash_value

    def copy(self):
        center, dir_x, dir_y, dir_z = self.get_params()
        return BOX(center, dir_x, dir_y, dir_z, **self.options)


class Plane(Surface, _Plane):
    """Plane surface class.

    Parameters
    ----------
    normal : array_like[float]
        The normal to the plane being created.
    offset : float
        Free term.
    options : dict
        Dictionary of surface's options. Possible values:
            transform = tr - transformation to be applied to this plane.
                             Transformation instance.
    """

    def __init__(self, normal: VectorLike, offset: float, assume_normalized=False, **options: Any):
        v = np.asarray(normal, dtype=float)
        k = float(offset)
        if not assume_normalized:
            v, is_ort = internalize_ort(v)
            if not is_ort:
                length = np.linalg.norm(v)
                v = v / length
                k /= length
        Surface.__init__(self, **options)
        _Plane.__init__(self, v, k)
        self._hash = compute_hash(self._k, self._v, self.transformation)

    # noinspection PyTypeChecker
    def round(self):
        result = self.apply_transformation()
        k_digits = significant_digits(
            result._k, constants.FLOAT_TOLERANCE, constants.FLOAT_TOLERANCE
        )
        v_digits = significant_array(
            result._v, constants.FLOAT_TOLERANCE, constants.FLOAT_TOLERANCE
        )
        k = round_scalar(result._k, k_digits)
        v = round_array(result._v, v_digits)
        return Plane(v, k, transform=None, assume_normalized=True, **result.options)

    def copy(self):
        options = filter_dict(self.options)
        instance = Plane(self._v, self._k, assume_normalized=True, **options)
        return instance

    __deepcopy__ = copy

    def apply_transformation(self) -> "Plane":
        tr = self.transformation
        if tr is None:
            return self
        v, k = tr.apply2plane(self._v, self._k)
        options = self.clean_options()
        return Plane(v, k, transform=None, assume_normalized=True, **options)

    def transform(self, tr: Transformation) -> "Plane":
        if tr is None:
            return self
        tr = self.combine_transformations(tr)
        options = self.clean_options()
        return Plane(self._v, self._k, transform=tr, assume_normalized=True, **options)

    def reverse(self):
        """Gets the surface with reversed normal."""
        options = self.clean_options()
        return Plane(-self._v, -self._k, assume_normalized=True, **options)

    def mcnp_words(self, pretty: bool = False) -> List[str]:
        words = Surface.mcnp_words(self, pretty)
        if np.array_equal(self._v, EX):
            words.append("PX")
        elif np.array_equal(self._v, EY):
            words.append("PY")
        elif np.array_equal(self._v, EZ):
            words.append("PZ")
        else:
            words.append("P")
            for v in self._v:
                add_float(words, v, pretty)
        add_float(
            words, -self._k, pretty
        )  # TODO dvp: check why is offset negated in create_surface()?
        return words

    def is_close_to(
        self,
        other: "Surface",
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        if self is other:
            return True
        if not isinstance(other, Plane):
            return False
        return estimator(
            (self._k, self._v, self.transformation),
            (other._k, other._v, other.transformation),
        )

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Plane):
            return False
        return are_equal(
            (self._k, self._v, self.transformation),
            (other._k, other._v, other.transformation),
        )

    def __getstate__(self):
        return self._v, self._k, Surface.__getstate__(self)

    def __setstate__(self, state):
        v, k, options = state
        self.__init__(v, k, assume_normalized=True, **options)

    def __repr__(self):
        return f"Plane({self._v}, {self._k}, {self.options if self.options else ''})"


# noinspection PyProtectedMember,PyTypeChecker
class Sphere(Surface, _Sphere):
    """Sphere surface class.

    Parameters
    ----------
    center : array_like[float]
        Center of the sphere.
    radius : float
        The radius of the sphere.
    options : dict
        Dictionary of surface's options. Possible values:
            transform = tr - transformation to be applied to the sphere being
                             created. Transformation instance.
    """

    def __init__(self, center: Union[Sequence[float], np.ndarray], radius: float, **options: Any):
        center = np.asarray(center, dtype=float)
        radius = float(radius)
        Surface.__init__(self, **options)
        _Sphere.__init__(self, center, radius)
        self._hash = compute_hash(self._radius, self._center, self.transformation)

    def __getstate__(self):
        return self._center, self._radius, Surface.__getstate__(self)

    def __setstate__(self, state):
        c, r, options = state
        self.__init__(c, r, **options)

    def copy(self):
        return Sphere(self._center, self._radius, **deepcopy(self.options))

    __deepcopy__ = copy

    def __repr__(self):
        return f"Sphere({self._center}, {self._radius}, {self.options if self.options else ''})"

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Sphere):
            return False  # TODO dvp: what if `other` is GQuadratic representation of Sphere?
        return are_equal(
            (self._radius, self._center, self.transformation),
            (other._radius, other._center, other.transformation),
        )

    def is_close_to(
        self,
        other: "Sphere",
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        if self is other:
            return True
        if not isinstance(other, Sphere):
            return False
        return estimator(
            (self._radius, self._center, self.transformation),
            (other._radius, other._center, other.transformation),
        )

    def round(self) -> "Surface":
        temp = self.apply_transformation()
        center_digits = significant_array(
            temp._center,
            constants.FLOAT_TOLERANCE,
            resolution=constants.FLOAT_TOLERANCE,
        )
        center = round_array(temp._center, center_digits)
        radius_digits = significant_digits(
            temp._radius,
            constants.FLOAT_TOLERANCE,
            resolution=constants.FLOAT_TOLERANCE,
        )
        radius = round_scalar(temp._radius, radius_digits)
        options = temp.clean_options()
        return Sphere(center, radius, transform=None, **options)

    def apply_transformation(self) -> "Sphere":
        tr = self.transformation
        if tr is None:
            return self
        center = tr.apply2point(self._center)
        options = self.clean_options()
        return Sphere(center, self._radius, transform=None, **options)

    def transform(self, tr: Transformation) -> "Sphere":
        if tr is None:
            return self
        tr = self.combine_transformations(tr)
        options = self.clean_options()
        return Sphere(self._center, self._radius, transform=tr, **options)

    def mcnp_words(self, pretty: bool = False) -> List[str]:
        words = Surface.mcnp_words(self, pretty)
        c = self._center
        if np.array_equal(self._center, ORIGIN):
            words.append("SO")
        elif c[0] == 0.0 and c[1] == 0.0:
            words.append("SZ")
            add_float(words, c[2], pretty)
        elif c[1] == 0.0 and c[2] == 0.0:
            words.append("SX")
            add_float(words, c[0], pretty)
        elif c[0] == 0.0 and c[2] == 0.0:
            words.append("SY")
            add_float(words, c[1], pretty)
        else:
            words.append("S")
            for v in c:
                add_float(words, v, pretty)
        add_float(words, self._radius, pretty)
        return words


# noinspection PyProtectedMember,PyUnresolvedReferences,DuplicatedCode,PyTypeChecker
class Cylinder(Surface, _Cylinder):
    """Cylinder surface class.

    Parameters
    ----------
    pt : array_like[float]
        Point on the cylinder's axis.
    axis : array_like[float]
        Cylinder's axis direction.
    radius : float
        Cylinder's radius.
    options : dict
        Dictionary of surface's options. Possible values:
            transform = tr - transformation to be applied to the cylinder being
                             created. Transformation instance.
    """

    def __init__(self, pt, axis, radius, assume_normalized=False, **options):
        axis = np.asarray(axis, dtype=float)
        if not assume_normalized:
            axis, is_ort = internalize_ort(axis)
            if not is_ort:
                axis /= np.linalg.norm(axis)
        max_dir = np.argmax(np.abs(axis))
        if axis[max_dir] < 0:
            axis *= -1
        pt = np.asarray(pt, dtype=float)
        pt = pt - axis * np.dot(pt, axis)
        Surface.__init__(self, **options)
        _Cylinder.__init__(self, pt, axis, radius)
        self._hash = compute_hash(self._radius, self._pt, self._axis, self.transformation)

    def copy(self):
        return Cylinder(
            self._pt,
            self._axis,
            self._radius,
            assume_normalized=True,
            **deepcopy(self.options),
        )

    def __repr__(self):
        return f"Cylinder({self._pt}, {self._axis}, {self.options if self.options else ''})"

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Cylinder):
            return False
        return are_equal(
            (self._radius, self._pt, self._axis, self.transformation),
            (other._radius, other._pt, other._axis, other.transformation),
        )

    def is_close_to(
        self,
        other: "Cylinder",
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        if self is other:
            return True
        if not isinstance(other, Cone):
            return False
        return estimator(
            (self._radius, self._pt, self._axis, self.transformation),
            (other._radius, other._pt, other._axis, other.transformation),
        )

    def round(self):
        temp = self.apply_transformation()
        pt = round_array(temp._pt)
        axis = round_array(temp._axis)
        radius = round_scalar(temp._radius)
        options = self.clean_options()
        return Cylinder(pt, axis, radius, transform=None, **options)

    def apply_transformation(self) -> Surface:
        tr = self.transformation
        if tr is None:
            return self
        pt = tr.apply2point(self._pt)
        axis = tr.apply2vector(self._axis)
        options = self.clean_options()
        # TODO dvp: actually may create Generic Quadratic. Should we use __new__() for this?
        return Cylinder(pt, axis, self._radius, assume_normalized=True, **options)

    def transform(self, tr: Transformation) -> "Cylinder":
        if tr is None:
            return self
        tr = self.combine_transformations(tr)
        options = self.clean_options()
        return Cylinder(
            self._pt,
            self._axis,
            self._radius,
            transform=tr,
            assume_normalized=True,
            **options,
        )

    def mcnp_words(self, pretty: bool = False) -> str:
        words = Surface.mcnp_words(self)
        axis = self._axis
        pt = self._pt
        if np.array_equal(axis, EX):
            if pt[1] == 0.0 and pt[2] == 0.0:
                words.append("CX")
            else:
                words.append("C/X")
                add_float(words, pt[1], pretty)
                add_float(words, pt[2], pretty)
        elif np.array_equal(axis, EY):
            if pt[0] == 0.0 and pt[2] == 0.0:
                words.append("CY")
            else:
                words.append("C/Y")
                add_float(words, pt[0], pretty)
                add_float(words, pt[2], pretty)
        elif np.array_equal(axis, EZ):
            if pt[0] == 0.0 and pt[1] == 0.0:
                words.append("CZ")
            else:
                words.append("C/Z")
                add_float(words, pt[0], pretty)
                add_float(words, pt[1], pretty)
        else:
            nx, ny, nz = axis
            m = np.array(
                [
                    [1 - nx**2, -nx * ny, -nx * nz],
                    [-nx * ny, 1 - ny**2, -ny * nz],
                    [-nx * nz, -ny * nz, 1 - nz**2],
                ],
                dtype=float,
            )
            v = np.zeros(3)
            k = -self._radius**2
            m, v, k = Transformation(translation=self._pt).apply2gq(m, v, k)
            return GQuadratic(m, v, k, **self.options).mcnp_repr(pretty)
        add_float(words, self._radius, pretty)
        return words

    def __getstate__(self):
        return self._pt, self._axis, self._radius, Surface.__getstate__(self)

    def __setstate__(self, state):
        pt, axis, radius, options = state
        self.__init__(pt, axis, radius, assume_normalized=True, **options)


# noinspection PyProtectedMember,PyUnresolvedReferences,DuplicatedCode,PyTypeChecker
class Cone(Surface, _Cone):
    """Cone surface class.

    Parameters
    ----------
    apex : array_like[float]
        Cone's apex.
    axis : array_like[float]
        Cone's axis.
    t2 : float
        Square of tangent of angle between axis and generatrix.
    sheet : int
        Cone's sheet.
    options : dict
        Dictionary of surface's options. Possible values:
            transform = tr - transformation to be applied to the cone being
                             created. Transformation instance.
    """

    def __init__(
        self,
        apex: VectorLike,
        axis: VectorLike,
        t2: float,
        sheet: int = 0,
        assume_normalized: bool = False,
        **options,
    ) -> None:
        axis = np.asarray(axis, dtype=float)
        if not assume_normalized:
            axis, is_ort = internalize_ort(axis)
            if not is_ort:
                axis /= np.linalg.norm(axis)
        maxdir = np.argmax(np.abs(axis))
        if axis[maxdir] < 0:
            axis *= -1
            sheet *= -1
        apex = np.asarray(apex, dtype=float)
        Surface.__init__(self, **options)
        _Cone.__init__(self, apex, axis, t2, sheet)
        self._hash = compute_hash(self._t2, self._sheet, self._apex, self._axis)

    def apply_transformation(self) -> Surface:
        tr = self.transformation
        if tr is None:
            return self
        apex = tr.apply2point(self._apex)
        axis = tr.apply2vector(self._axis)
        sheet = self._sheet
        options = self.clean_options()
        return Cone(apex, axis, self._t2, sheet, assume_normalized=True, **options)

    def round(self) -> Surface:
        res = self.apply_transformation()
        apex = round_array(res._apex)
        axis = round_array(res._axis)
        t2 = round_scalar(self._t2)
        sheet = self._sheet
        options = self.clean_options()
        # TODO dvp: what if Generic Quadratic is to be returned?
        return Cone(apex, axis, t2, sheet, assume_normalized=True, **options)

    def copy(self):
        return Cone(
            self._apex,
            self._axis,
            self._t2,
            self._sheet,
            assume_normalized=True,
            **deepcopy(self.options),
        )

    def __repr__(self):
        return f"Cone({self._apex}, {self._axis}, {self._t2}, {self._sheet}, {self.options if self.options else ''})"

    def __getstate__(self):
        return self._apex, self._axis, self._t2, self._sheet, Surface.__getstate__(self)

    def __setstate__(self, state):
        apex, axis, t2, sheet, options = state
        self.__init__(apex, axis, t2, sheet, **options)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):

        # noinspection DuplicatedCode
        if self is other:
            return True

        if not isinstance(other, Cone):
            return False

        return are_equal(
            (self._t2, self._sheet, self._apex, self._axis, self.transformation),
            (other._t2, other._sheet, other._apex, other._axis, self.transformation),
        )

    def is_close_to(
        self,
        other: "Surface",
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        if self is other:
            return True
        if not isinstance(other, Cone):
            return False
        return estimator(
            (self._apex, self._axis, self._t2, self.transformation),
            (other._apex, other._axis, other._t2, other.transformation),
        )

    def transform(self, tr: Transformation) -> "Surface":
        if tr is None:
            return self
        tr = self.combine_transformations(tr)
        options = self.clean_options()
        cone = Cone(self._apex, self._axis, self._t2, sheet=self._sheet, transform=tr, **options)
        return cone

    def mcnp_words(self, pretty: bool = False) -> List[str]:
        words = Surface.mcnp_words(self)
        axis = self._axis
        apex = self._apex
        if np.array_equal(axis, EX):
            if apex[1] == 0.0 and apex[2] == 0.0:
                words.append("KX")
                add_float(words, apex[0], pretty)
            else:
                words.append("K/X")
                for v in apex:
                    add_float(words, v, pretty)
        elif np.array_equal(axis, EY):
            if apex[0] == 0.0 and apex[2] == 0.0:
                words.append("KY")
                add_float(words, apex[1], pretty)
            else:
                words.append("K/Y")
                for v in apex:
                    add_float(words, v, pretty)
        elif np.array_equal(axis, EZ):
            if apex[0] == 0.0 and apex[1] == 0.0:
                words.append("KZ")
                add_float(words, apex[2], pretty)
            else:
                words.append("K/Z")
                for v in apex:
                    add_float(words, v, pretty)
        else:
            nx, ny, nz = axis
            a = 1 + self._t2
            m = np.array(
                [
                    [1 - a * nx**2, -a * nx * ny, -a * nx * nz],
                    [-a * nx * ny, 1 - a * ny**2, -a * ny * nz],
                    [-a * nx * nz, -a * ny * nz, 1 - a * nz**2],
                ]
            )
            v = np.zeros(3)
            k = 0
            m, v, k = Transformation(translation=self._apex).apply2gq(m, v, k)
            return GQuadratic(m, v, k, **self.clean_options()).mcnp_repr(pretty)
        add_float(words, self._t2, pretty)
        if self._sheet != 0:
            words.append(" ")
            words.append(str(self._sheet))
        return words


# noinspection PyProtectedMember,PyUnresolvedReferences,PyTypeChecker
class GQuadratic(Surface, _GQuadratic):
    """Generic quadratic surface class.

    Parameters
    ----------
    m : array_like[float]
        Matrix of coefficients of quadratic terms. m.shape=(3,3)
    v : array_like[float]
        Vector of coefficients of linear terms. v.shape=(3,)
    k : float
        Free term.
    options : dict
        Dictionary of surface's options. Possible values:
            transform = tr - transformation to be applied to the surface being
                             created. Transformation instance.
    """

    def __init__(self, m, v, k, factor=None, **options):
        m = np.asarray(m, dtype=float)
        v = np.asarray(v, dtype=float)
        if factor is None:
            eigenvalues = np.linalg.eigvalsh(m)
            factor = 1.0 / np.max(np.abs(eigenvalues))
        Surface.__init__(self, **options)
        _GQuadratic.__init__(self, m, v, k, factor)
        self._hash = compute_hash(self._k, self._v, self._m)

    def apply_transformation(self) -> Surface:
        tr = self.transformation
        if tr is None:
            return self
        m, v, k = tr.apply2gq(self._m, self._v, self._k)
        options = self.clean_options()
        return GQuadratic(m, v, k, **options)

    def round(self) -> Surface:
        temp: Surface = self.apply_transformation()
        m, v = map(round_array, [temp._m, temp._v])
        k = round_scalar(temp._k)
        # TODO dvp: handle cases when the surface can be represented with specialized quadratic surface: Cone etc.
        return GQuadratic(m, v, k, **self.clean_options())

    def __getstate__(self):
        return self._m, self._v, self._k, self._factor, Surface.__getstate__(self)

    def __setstate__(self, state):
        m, v, k, factor, options = state
        self.__init__(m, v, k, factor, **options)

    def copy(self) -> "GQuadratic":
        options = deepcopy(self.options)
        return GQuadratic(self._m, self._v, self._k, self._factor, **options)

    def __repr__(self):
        options = str(self.options) if self.options else ""
        return f"GQuadratic({self._m}, {self._v}, {self._k}, {self._factor}, {options})"

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other) -> bool:
        if self is other:
            return True

        if not isinstance(other, GQuadratic):
            return False
        # TODO dvp: check if `self.factor` is to be accounted to as well.
        return are_equal(
            (self._k, self._v, self._m, self.transformation),
            (other._k, other._v, other._m, other.transformation),
        )

    def is_close_to(
        self,
        other: Surface,
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        if self is other:
            return True
        if not isinstance(other, GQuadratic):
            return False  # TODO dvp: handle cases when other is specialized quadratic surface Sphere, Cone etc.
        return estimator(
            (self._k, self._v, self._m, self.transformation),
            (other._k, other._v, other._m, other.transformation),
        )

    def transform(self, tr):
        if tr is None:
            return self
        tr = self.combine_transformations(tr)
        options = self.clean_options()
        return GQuadratic(self._m, self._v, self._k, transform=tr, **options)

    def mcnp_words(self, pretty: bool = False) -> List[str]:
        words = Surface.mcnp_words(self)
        words.append("GQ")
        m = self._m
        a, b, c = np.diag(m)
        d = m[0, 1] + m[1, 0]
        e = m[1, 2] + m[2, 1]
        f = m[0, 2] + m[2, 0]
        g, h, j = self._v
        k = self._k
        for v in [a, b, c, d, e, f, g, h, j, k]:
            add_float(words, v, pretty)
        return words


# noinspection PyProtectedMember,PyUnresolvedReferences,PyTypeChecker
class Torus(Surface, _Torus):
    """Tori surface class.

    Parameters
    ----------
    center : array_like[float]
        The center of torus.
    axis : array_like[float]
        The axis of torus.
    r : float
        Major radius.
    a : float
        Radius parallel to torus axis.
    b : float
        Radius perpendicular to torus axis.
    options : dict
        Dictionary of surface's options. Possible values:
            transform = tr - transformation to be applied to the torus being
                             created. Transformation instance.
    """

    def __init__(self, center, axis, r, a, b, assume_normalized=False, **options):
        center = np.asarray(center, dtype=float)
        axis = np.asarray(axis, dtype=float)
        if not assume_normalized:
            axis, is_ort = internalize_ort(axis)
            if not is_ort:
                axis /= np.linalg.norm(axis)
        maxdir = np.argmax(np.abs(axis))
        if axis[maxdir] < 0:
            axis *= -1
        Surface.__init__(self, **options)
        _Torus.__init__(self, center, axis, r, a, b)
        self._hash = compute_hash(self._center, self._axis, self._R, self._a, self._b)

    def round(self) -> "Surface":
        temp = self.apply_transformation()
        center, axis = map(round_array, [temp._center, temp._axis])

        def r(x):
            return round_scalar(x, significant_digits(x, FLOAT_TOLERANCE))

        r, a, b = map(r, [temp._R, temp._a, temp._b])
        options = temp.clean_options()
        return Torus(center, axis, r, a, b, assume_normalized=True, **options)

    def apply_transformation(self) -> "Surface":
        tr = self.transformation
        if tr is None:
            return self
        center = tr.apply2point(self._center)
        axis = tr.apply2vector(self._axis)
        # TODO dvp: should we check the transformation and result? The axis is to be along EX, EY, EZ.
        return Torus(
            center,
            axis,
            self._R,
            self._a,
            self._b,
            assume_normalized=True,
            **self.clean_options(),
        )

    def __getstate__(self):
        return (
            self._center,
            self._axis,
            self._R,
            self._a,
            self._b,
            Surface.__getstate__(self),
        )

    def __setstate__(self, state):
        center, axis, r, a, b, options = state
        self.__init__(center, axis, r, a, b, **options)

    def copy(self):
        return Torus(
            self._center,
            self._axis,
            self._R,
            self._a,
            self._b,
            assume_normalized=True,
            **deepcopy(self.options),
        )

    def __hash__(self):
        return self._hash

    def __eq__(self, other: "Torus"):
        if self is other:
            return True
        if not isinstance(other, Torus):
            return False
        return are_equal(
            (self._center, self._axis, self._R, self._a, self._b, self.transformation),
            (
                other._center,
                other._axis,
                other._R,
                other._a,
                other._b,
                other.transformation,
            ),
        )

    def is_close_to(
        self,
        other: "Torus",
        estimator: Callable[[Any, Any], bool] = DEFAULT_TOLERANCE_ESTIMATOR,
    ) -> bool:
        if self is other:
            return True
        if not isinstance(other, Torus):
            return False
        return estimator(
            (self._center, self._axis, self._R, self._a, self._b, self.transformation),
            (
                other._center,
                other._axis,
                other._R,
                other._a,
                other._b,
                other.transformation,
            ),
        )

    def transform(self, tr):
        if tr is None:
            return self
        tr = self.combine_transformations(tr)
        options = self.clean_options()
        return Torus(self._center, self._axis, self._R, self._a, self._b, transform=tr, **options)

    def mcnp_words(self, pretty: bool = False) -> List[str]:
        words = Surface.mcnp_words(self)
        estimator = DEFAULT_TOLERANCE_ESTIMATOR
        if estimator(self._axis, EX):
            words.append("TX")
        elif estimator(self._axis, EY):
            words.append("TY")
        elif estimator(self._axis, EZ):
            words.append("TZ")
        else:
            raise NotImplementedError("The axis of a torus should be along EX, EY or EZ")
        x, y, z = self._center
        for v in [x, y, z, self._R, self._a, self._b]:
            add_float(words, v, pretty)
        return words

    def __repr__(self):
        return f"Torus({self._center}, {self._axis}, {self._R}, \
            {self._a}, {self._b}, {self.options if self.options else ''}"



from typing import Generator, Iterable, List, Optional, TextIO, Tuple

import re
import sys

from enum import IntEnum

from attr import attrib, attrs

BLANK_LINE_PATTERN = re.compile(r"\n\s*\n", flags=re.MULTILINE)
COMMENT_LINE_PATTERN = re.compile(r"^\s{,5}[cC]( .*)?\s*$")
# pattern to remove comments from a card text
REMOVE_COMMENT_PATTERN = re.compile(
    r"(\s*\$.*$)|(^\s{0,5}c\s.*\n?)", flags=re.MULTILINE | re.IGNORECASE
)
# pattern to split section text to optional "comment" and subsequent MCNP card pairs
CARD_PATTERN = re.compile(
    r"(?P<comment>((^\s{,5}c( .*)?\s*$)\n?)+)?(?P<card>^\s{,5}(\*|\+|\w).*(\n((^\s{,5}c.*\n?)*^\s{5,}\S.*\n?)*)?)?",
    flags=re.MULTILINE | re.IGNORECASE,
)

# pattern to replace subsequent spaces with a single one
SPACE_PATTERN = re.compile(r"\s+", flags=re.MULTILINE)

SDEF_PATTERN = re.compile(r"(sdef)|(s[ibp]\d+)|(ds\d+)|(wwp.*)", re.IGNORECASE)

# pattern to match labels of cards relevant for tallies section
TALLY_PATTERN = re.compile(r"(?P<tag>f(c|(m(esh)?)?)|(e)|(d[ef]))(?P<number>\d+)", re.IGNORECASE)


class Kind(IntEnum):
    COMMENT = 0
    CELL = 1
    SURFACE = 2
    MATERIAL = 3
    TRANSFORMATION = 4
    SDEF = 5
    TALLY = 6
    GENERIC = 7

    @classmethod
    def from_card_text(cls, text: str):
        label: str = text.split(maxsplit=2)[0]
        label = label.strip()
        if label[0] in "mM" and str.isdigit(label[1]):
            return Kind.MATERIAL
        # check if a label is for transformation card, example: TR20 or *tr1
        i = 0
        if label[0] == "*":
            i = 1
        if label[i] in "tT" and label[i + 1] in "rR" and str.isdigit(label[i + 2]):
            return Kind.TRANSFORMATION
        if SDEF_PATTERN.match(label):
            return Kind.SDEF
        if TALLY_PATTERN.match(label):
            return Kind.TALLY
        if is_comment(text):
            return Kind.COMMENT
        return Kind.GENERIC


@attrs(str=False)
class Card(object):
    """
    Generic MCNP card raw text item

    The card kind is defined either from parsing context (for cells and surfaces) or
    from card label
    """

    text: str = attrib()
    kind: Optional[Kind] = attrib(default=None)

    # noinspection PyUnusedLocal,PyUnresolvedReferences
    @kind.validator
    def check(self, attribute, value) -> None:
        if self.kind is None:
            self.kind = Kind.from_card_text(self.text)

    @property
    def is_comment(self) -> bool:
        return self.kind is Kind.COMMENT

    @property
    def is_material(self) -> bool:
        return self.kind is Kind.MATERIAL

    @property
    def is_transformation(self) -> bool:
        return self.kind is Kind.TRANSFORMATION

    @property
    def is_cell(self) -> bool:
        return self.kind is Kind.CELL

    @property
    def is_surface(self) -> bool:
        return self.kind is Kind.SURFACE

    @property
    def is_sdef(self) -> bool:
        return self.kind is Kind.SDEF

    @property
    def is_tally(self) -> bool:
        return self.kind is Kind.TALLY

    def get_clean_text(self) -> str:
        return get_clean_text(self.text)

    def __str__(self) -> str:
        return self.text

    @property
    def name(self) -> str:
        assert not self.is_comment, "Comment cannot have name"
        return self.text.split(maxsplit=2)[0]

    @property
    def number(self) -> int:
        name = self.name
        if self.is_cell:
            return int(name)
        if self.is_surface:
            if name.isdigit():
                return int(name)
            else:
                assert name[0] in "%*", "Expected reflecting, white surface"
                return int(name[1:])
        if self.is_material:
            return int(name[1:])
        if self.is_transformation:
            i = 0
            if name[0] == "*":
                i = 1
            if name[i] in "tT" and name[i + 1] in "rR" and name[i + 2 :].isdigit():
                return int(name[i + 2 :])
        raise NotImplementedError(f"Cannot define number for {self.kind} card '{name}'")


@attrs
class InputSections(object):
    title: Optional[str] = attrib(default=None)
    cell_cards: Optional[List[Card]] = attrib(default=None)
    surface_cards: Optional[List[Card]] = attrib(default=None)
    data_cards: Optional[List[Card]] = attrib(default=None)
    message: Optional[str] = attrib(default=None)
    remainder: Optional[str] = attrib(default=None)
    is_continue: bool = attrib(default=False)

    # noinspection PyUnusedLocal,PyUnresolvedReferences
    @is_continue.validator
    def check(self, attribute, value) -> None:
        if self.is_continue:
            if self.cell_cards or self.surface_cards:
                raise ValueError("Cells and Surfaces shouldn't present in 'continue' mode model")
            if not self.data_cards:
                raise ValueError("At least one data card should present in 'continue' mode model")

    def print(self, stream=sys.stdout):
        if self.message:
            print(self.message, file=stream)
            print(file=stream)
        if self.title:
            print(self.title, file=stream)
        if not self.is_continue:
            for card in self.cell_cards:
                print(card.text, file=stream)
            print(file=stream)
            for card in self.surface_cards:
                print(card.text, file=stream)
            print(file=stream)
        if self.data_cards:
            for card in self.data_cards:
                print(card.text, file=stream)
            print(file=stream)
        if self.remainder:
            print(self.remainder, file=stream)


def split_to_cards(text: str, kind: Kind = None) -> Generator[Card, None, None]:
    for res in CARD_PATTERN.finditer(text):
        groups = res.groupdict()
        comment = groups["comment"]
        card = groups["card"]
        if comment:
            comment = comment.rstrip()
            yield Card(comment, kind=Kind.COMMENT)
        if card:
            card = card.rstrip()
            if kind is None:
                _kind = Kind.from_card_text(card)
            else:
                _kind = kind
            yield Card(card, kind=_kind)


def parse_sections(inp: TextIO) -> InputSections:
    return parse_sections_text(inp.read())


# noinspection PyIncorrectDocstring,PyUnresolvedReferences
def parse_sections_text(text: str) -> InputSections:
    """
    Splits input file to sections according to mcnp input format.

    Long Description
    ----------------
    The parser provides minimal separation of text sections from the input file
    based on Blank Line Delimiter entries
    The sections can be parsed later and in order different from the input file
    using specific low level parsers for cell, surface and data cards.

    From MCNP manual
    ~~~~~~~~~~~~~~~~
    An input file has the following form::
        Message Block           | Optional
        Blank Line Delimiter    | Required if Message Block is present
        Title Card
        Cell Cards
        Blank Line Delimiter
        Surface Cards
        Blank Line Delimiter
        Data Cards
        Blank Line Terminator    | Recommended
        Anything Else            | Optional


    Parameters
    ----------
    inp - stream

    Exceptions
    ----------
    The parser performs verification of the input using builder object,
    on errors the builder raises SectionParserError.

    Returns
    -------
    mcnp_input_sections object, containing text of sections from the file
    """

    message = None
    cell_cards = None
    surface_cards = None
    data_cards = None
    remainder = None

    sections = BLANK_LINE_PATTERN.split(text, 5)

    # The first line can be message or title.
    kw = sections[0][: len("message:")].lower()

    i = 0

    if "message:" == kw:
        message = sections[0]
        i += 1

    title, cur_section = sections[i].split("\n", 1)
    i += 1
    if not title:
        raise ValueError("Cannot find the MCNP model title")

    is_continue = check_title_is_continue(title)

    if is_continue:
        data_cards = list(split_to_cards(cur_section))
        if i < len(sections):
            remainder = "\n\n".join(sections[i:])
        result = InputSections(
            title,
            cell_cards=cell_cards,
            surface_cards=surface_cards,
            data_cards=data_cards,
            message=message,
            remainder=remainder,
            is_continue=True,
        )
    else:
        cell_cards = list(split_to_cards(cur_section, kind=Kind.CELL))
        if i < len(sections):
            cur_section = sections[i]
            surface_cards = list(split_to_cards(cur_section, kind=Kind.SURFACE))
            i += 1
            if i < len(sections):
                cur_section = sections[i]
                data_cards = list(split_to_cards(cur_section, kind=None))
                i += 1
                if i < len(sections):
                    remainder = "\n\n".join(sections[i:])
        result = InputSections(
            title,
            cell_cards=cell_cards,
            surface_cards=surface_cards,
            data_cards=data_cards,
            message=message,
            remainder=remainder,
            is_continue=False,
        )

    return result


CONTINUE_LEN = len("continue")


def check_title_is_continue(title):
    return title[:CONTINUE_LEN].lower() == "continue"


def is_comment(text: str) -> bool:
    assert isinstance(text, str), "The parameter 'line' should be text"
    if "\n" in text:
        res = next((line for line in text.split("\n") if not is_comment_line(line)), False)
        return not res
    else:
        return is_comment_line(text, skip_asserts=True)


def is_comment_line(line: str, skip_asserts=False) -> bool:
    if not skip_asserts:
        assert isinstance(line, str), "The parameter 'line' should be text"
        assert "\n" not in line, "The parameter 'line' should be the single text line"
    return COMMENT_LINE_PATTERN.match(line) is not None


def get_clean_text(text: str) -> str:
    without_comments = REMOVE_COMMENT_PATTERN.sub("", text)
    with_spaces_normalized = SPACE_PATTERN.sub(" ", without_comments)
    return with_spaces_normalized


def clean_mcnp_cards(iterable: Iterable[Card]) -> Generator[Card, None, None]:
    for x in iterable:
        if x.kind is not Kind.COMMENT:
            clean_text = x.get_clean_text()
            t = Card(clean_text, kind=x.kind)
            yield t


def distribute_cards(
    cards: Iterable[Card],
) -> Tuple[List[Card], List[Card], List[Card], List[Card], List[Card]]:
    comment: Optional[Card] = None

    def append(_cards: List[Card], _card: Card) -> None:
        nonlocal comment
        if comment:
            _cards.append(comment)
            comment = None
        _cards.append(_card)

    # fmt: off
    materials, transformations, sdef, tallies, others = \
        [], [], [], [], []  # type: List[Card], List[Card], List[Card], List[Card], List[Card]
    # fmt: on

    for card in cards:
        if card.is_comment:
            assert comment is None
            comment = card
        elif card.is_material:
            append(materials, card)
        elif card.is_transformation:
            append(transformations, card)
        elif card.is_sdef:
            append(sdef, card)
        elif card.is_tally:
            append(tallies, card)
        else:
            append(others, card)

    if comment:
        others.append(comment)

    return materials, transformations, sdef, tallies, others



# -*- coding: utf-8 -*-
"""Apply transformation to a model."""
from pathlib import Path

from mckit.cli.logging import logger
from mckit.parser.mcnp_input_sly_parser import ParseResult, from_file
from mckit.parser.mcnp_section_parser import clean_mcnp_cards, split_to_cards
from mckit.parser.transformation_parser import parse as parse_transformation
from mckit.universe import Universe
from mckit.utils.indexes import IndexOfNamed, raise_on_duplicate_strategy

from .common import save_mcnp


def transform(
    output: Path,
    transformation: int,
    transformations: Path,
    source: Path,
    override: bool,
) -> None:
    logger.info("Running mckit transform")
    logger.debug("Working dir {}", Path(".").absolute())
    logger.info("Transforming model from {s}", s=source)
    if output.exists() and not override:
        raise FileExistsError(f"File {output} already exists. Remove it or use --override option")
    parse_result: ParseResult = from_file(source)
    src: Universe = parse_result.universe
    logger.debug("Loading transformations {} from {}", transformation, transformations)
    transformations_text = transformations.read_text()
    transformations_list = [
        parse_transformation(c.text) for c in clean_mcnp_cards(split_to_cards(transformations_text))
    ]
    transformations_index = IndexOfNamed.from_iterable(
        transformations_list,
        on_duplicate=raise_on_duplicate_strategy,
    )
    if transformation not in transformations_index:
        raise ValueError(f"Transformation {transformation} is not found in {transformations}")
    the_transformation = transformations_index[transformation]
    dst = src.transform(the_transformation)
    save_mcnp(dst, output, override)
    logger.success("Transformed model is saved to {}", output)



import mckit.parser.transformation_parser as trp
import pytest

from mckit.transformation import Transformation


@pytest.mark.parametrize(
    "text,expected_types,expected_values",
    [
        (
            "tr1 0 0 1.0",
            ["NAME", "INTEGER", "INTEGER", "FLOAT"],
            [(1, False), 0, 0, 1.0],
        ),
        (
            "*tr1 0 0 1.0",
            ["NAME", "INTEGER", "INTEGER", "FLOAT"],
            [(1, True), 0, 0, 1.0],
        ),
    ],
)
def test_transformation_lexer(text, expected_types, expected_values):
    lexer = trp.Lexer()
    tokens = [t for t in lexer.tokenize(text)]
    actual_types = [t.type for t in tokens]
    assert actual_types == expected_types
    actual_values = [t.value for t in tokens]
    assert actual_values == expected_values


# noinspection PyTypeChecker
@pytest.mark.parametrize(
    "text,expected",
    [
        ("tr2 0 0 1", Transformation(translation=[0.0, 0.0, 1.0], name=2)),
        (
            " *tr2 0 0 1 45 45 90 135 45 90 90 90 0",
            Transformation(
                translation=[0.0, 0.0, 1.0],
                rotation=[45, 45, 90, 135, 45, 90, 90, 90, 0],
                indegrees=True,
                name=2,
            ),
        ),
        (
            "*tr1 0. 0. 0. 3.62 86.38 90. 93.62 3.62 90. 90. 90. 0.",
            Transformation(
                translation=[0.0, 0.0, 0.0],
                rotation=[3.62, 86.38, 90.0, 93.62, 3.62, 90.0, 90.0, 90.0, 0.0],
                indegrees=True,
                name=1,
            ),
        ),
    ],
)
def test_transformation_parser(text, expected):
    actual = trp.parse(text)
    assert actual == expected


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""
 Read and parse MCNP file text.
"""
from typing import Any, Callable, Generator, Iterable, List, NewType, Optional, TextIO, Tuple, Union

from itertools import repeat
from pathlib import Path

from attr import attrib, attrs
from mckit.card import Card
from mckit.constants import MCNP_ENCODING
from mckit.parser.cell_parser import Body
from mckit.parser.cell_parser import parse as parse_cell
from mckit.parser.common import (
    CellNotFoundError,
    CellStrictIndex,
    CompositionStrictIndex,
    ParseError,
    SurfaceStrictIndex,
    TransformationStrictIndex,
)
from mckit.parser.material_parser import Composition
from mckit.parser.material_parser import parse as parse_composition
from mckit.parser.surface_parser import Surface
from mckit.parser.surface_parser import parse as parse_surface
from mckit.parser.transformation_parser import Transformation
from mckit.parser.transformation_parser import parse as parse_transformation
from mckit.universe import Universe, produce_universes
from mckit.utils.indexes import Index

from .mcnp_section_parser import Card as TextCard
from .mcnp_section_parser import InputSections, Kind, distribute_cards, parse_sections_text

T = NewType("T", Any)
T1 = NewType("T1", Any)
T2 = NewType("T2", Any)
Pair = NewType("Pair", Tuple[T1, T2])
YieldGenerator = NewType("YieldGenerator", Generator[T, None, None])


@attrs
class ParseResult:
    universe: Universe = attrib()
    cells: List[Body] = attrib()
    cells_index: CellStrictIndex = attrib()
    surfaces: List[Surface] = attrib()
    surfaces_index: SurfaceStrictIndex = attrib()
    compositions: Optional[List[Composition]] = attrib()
    compositions_index: Optional[CompositionStrictIndex] = attrib()
    transformations: Optional[List[Transformation]] = attrib()
    transformations_index: Optional[TransformationStrictIndex] = attrib()
    sections: InputSections = attrib()

    @property
    def title(self):
        return self.sections.title


def from_file(path: Union[str, Path]) -> ParseResult:
    if isinstance(path, str):
        path = Path(path)
    with path.open("r", encoding=MCNP_ENCODING) as fid:
        return from_stream(fid)


def from_stream(stream: TextIO) -> ParseResult:
    text = stream.read()
    return from_text(text)


def from_text(text: str) -> ParseResult:
    sections: InputSections = parse_sections_text(text)
    if sections.data_cards:
        # fmt: off
        text_compositions, text_transformations, _, _, _ = distribute_cards(
            sections.data_cards
        )  # type: List[TextCard], List[TextCard], List[TextCard], List[TextCard], List[TextCard],
        # fmt: on
        transformations = parse_transformations(text_transformations)
        transformations_index = TransformationStrictIndex.from_iterable(transformations)
        compositions = parse_compositions(text_compositions)
        compositions_index = CompositionStrictIndex.from_iterable(compositions)
    else:
        transformations = None
        transformations_index = None
        compositions = None
        compositions_index = None

    surfaces = parse_surfaces(sections.surface_cards, transformations_index)
    surfaces_index = SurfaceStrictIndex.from_iterable(surfaces)
    cells, cells_index = parse_cells(
        sections.cell_cards, surfaces_index, compositions_index, transformations_index
    )
    universe = produce_universes(
        cells
    )  # TODO dvp: apply title from sections to the topmost universe
    return ParseResult(
        universe=universe,
        cells=cells,
        cells_index=cells_index,
        surfaces=surfaces,
        surfaces_index=surfaces_index,
        compositions=compositions,
        compositions_index=compositions_index,
        transformations=transformations,
        transformations_index=transformations_index,
        sections=sections,
    )


TSectionGenerator = NewType(
    "TSectionGenerator",
    Generator[Union[Transformation, Composition, Surface], None, None],
)


def join_comments(text_cards: Iterable[TextCard]):
    def _iter():
        comment: Optional[str] = None
        for card in text_cards:
            if card.is_comment:
                assert comment is None, f"Comment is already set {comment[:70]}"
                comment = card.text
            else:
                assert (
                    not card.is_comment
                ), f"Pair of comment is found, second one is: {comment[:70]}"
                yield card, comment
                comment = None

    return list(_iter())


def parse_section(
    text_cards: Iterable[TextCard], expected_kind: Kind, parser: Callable[[str], Card]
) -> TSectionGenerator:
    text_cards_with_comments = join_comments(text_cards)

    def iterator() -> TSectionGenerator:
        for text_card, comment in text_cards_with_comments:
            assert text_card.kind is expected_kind
            try:
                card = parser(text_card.text)
            except (ValueError, ParseError) as ex:
                raise ValueError(f"Failed to parse card '{text_card}'") from ex
            if comment:
                card.options["comment_above"] = comment
            # card.options['original'] = text_card.text
            yield card

    return iterator()


def parse_transformations(text_cards: Iterable[TextCard]) -> List[Transformation]:
    return list(parse_section(text_cards, Kind.TRANSFORMATION, parse_transformation))


def parse_compositions(text_cards: Iterable[TextCard]) -> List[Composition]:
    return list(parse_section(text_cards, Kind.MATERIAL, parse_composition))


def parse_surfaces(text_cards: Iterable[TextCard], transformations: Index) -> List[Surface]:
    def parser(text: str):
        return parse_surface(text, transformations=transformations)

    return list(parse_section(text_cards, Kind.SURFACE, parser))


def extract_number(text_card: TextCard):
    return int(text_card.text.split(maxsplit=1)[0])


class MissedCellsError(RuntimeError):
    def __init__(self, missed_cells: List[int]):
        self.missed_cells = missed_cells
        msg = f"Not found cells: {missed_cells}"
        super().__init__(msg)

    @classmethod
    def from_text_cards(cls, missed_cells: Iterable[TextCard]):
        missed_cells_numbers: List[int] = list(map(extract_number, missed_cells))
        return cls(missed_cells_numbers)


def parse_cells(
    text_cards: List[TextCard],
    surfaces: Index,
    compositions: Index,
    transformations: Index,
) -> Tuple[List[Body], CellStrictIndex]:
    text_cards_with_comments = join_comments(text_cards)
    size = len(text_cards_with_comments)
    cells_index = CellStrictIndex()
    cells_to_process = list(range(size))
    cells = list(repeat(None, size))

    def parser(text: str):
        return parse_cell(
            text,
            cells=cells_index,
            surfaces=surfaces,
            compositions=compositions,
            transformations=transformations,
        )

    while cells_to_process:
        cells_to_process_length = len(cells_to_process)
        new_cells_to_process = []
        for i in cells_to_process:
            text_card, comment = text_cards_with_comments[i]
            assert text_card.kind is Kind.CELL
            try:
                card: Body = parser(text_card.text)
                assert card is not None, "Failed to process cell %s" % text_card.text[:70]
            except CellNotFoundError:
                new_cells_to_process.append(i)
                continue
            if comment:
                card.options["comment_above"] = comment
            card.options["original"] = text_card.text
            cells[i] = card
            cells_index[card.name()] = card
        if cells_to_process_length == len(new_cells_to_process):
            missed_cells_cards = list(text_cards[i] for i in new_cells_to_process)
            raise MissedCellsError.from_text_cards(missed_cells_cards)
        cells_to_process = new_cells_to_process

    if any(map(lambda c: c is None, cells)):
        cells = list(filter(lambda c: c is not None, cells))

    return cells, cells_index



import numpy as np
import pytest

from mckit.utils import (
    FLOAT_TOLERANCE,
    are_equal,
    compute_hash,
    filter_dict,
    get_decades,
    prettify_float,
    significant_digits,
)


@pytest.mark.parametrize(
    "value,expected",
    [
        (15.0, 1),
        (10.0, 1),
        (1.5, 0),
        (1.0, 0),
        (0.5, -1),
        (0.0011, -3),  # TODO dvp: check this strange values
        (0.001, -4),  #
        (0.01, -3),  #
        (0.1, -2),
        (1e-12, -13),
        (0.0, 0),
        (-1e-12, -13),
        (-0.01, -3),
        (-0.001, -4),
        (-0.0011, -3),
        (-0.5, -1),
        (-1.0, 0),
        (-1.5, 0),
        (-10.0, 1),
        (-15.0, 1),
    ],
)
def test_get_decades(value, expected):
    assert get_decades(value) == expected


@pytest.mark.parametrize(
    "value, reltol, resolution, expected",
    [
        (1.0, 1e-12, FLOAT_TOLERANCE, 0),
        (1.1, 1e-12, FLOAT_TOLERANCE, 1),
        (11.0, 1e-12, FLOAT_TOLERANCE, 0),
        (
            11.11,
            1e-12,
            FLOAT_TOLERANCE,
            2,
        ),  # TODO dvp: looks like the function should be called `digits_in_fraction`.
    ],
)
def test_significant_digits(value, reltol, resolution, expected):
    actual = significant_digits(value, reltol, resolution)
    assert actual == expected


@pytest.mark.parametrize(
    "a, b, expected",
    [
        (None, None, True),
        (None, "a", False),
        ("abc", "abc", True),
        (1, 1, True),
        (1, -1, False),
        (np.arange(10), np.arange(10), True),
        ([1, "a", np.arange(3)], [1, "a", np.arange(3)], True),
        ([1, "a", np.arange(3)], [2, "a", np.arange(3)], False),
        ([1, "a", np.arange(3)], [1, "a", np.arange(3), "a"], False),
    ],
)
def test_are_equal(a, b, expected):
    actual = are_equal(a, b)
    assert actual == expected


@pytest.mark.parametrize(
    "dictionary, drop_items, expected",
    [
        ({"a": 1, "b": 2}, "b", {"a": 1}),
        ({"a": 1, "b": 2, "c": 3}, frozenset(("b", "c")), {"a": 1}),
        ({"a": 1, "b": {"c": 3}}, "c", {"a": 1, "b": {}}),
        ({"a": 1, "b": {"c": 3}}, lambda x: x == "a", {"b": {"c": 3}}),
    ],
)
def test_deep_copy_dict(dictionary, drop_items, expected):
    actual = filter_dict(dictionary, drop_items)
    assert actual == expected


@pytest.mark.parametrize("values", [("abc",), ({1: {"a": 2}}, np.arange(10)), (None,)])
def test_make_hash(values):
    compute_hash(*values)


@pytest.mark.parametrize(
    "value, expected",
    [
        (1.0, "1"),
        (0.000000000001, "1e-12"),
        (0.00000000000101, "1.01e-12"),
        (5.00000000001, "5.00000000001"),
        (-2.828427124746, "-2.828427124746"),
    ],
)
def test_prettify_float(value, expected):
    actual = prettify_float(value)
    assert actual == expected



"""Generic utility methods."""
from __future__ import annotations

from typing import Any, Dict, Tuple, cast

import collections
import collections.abc
import functools
import itertools

from copy import deepcopy

import numpy as np

from numpy import ndarray

from ..constants import FLOAT_TOLERANCE

MAX_DIGITS = np.finfo(float).precision


def significant_digits(
    value: float, reltol: float = FLOAT_TOLERANCE, resolution: float = None
) -> int:
    """The minimum number of significant digits to provide relative tolerance.

    Args:
        value:  The value to be checked.
        reltol: Relative tolerance needed to represent the value.
        resolution:  The threshold value, below which numbers are believed to be zero, optional.

    Returns:
        The number of significant digits.
    """
    if value == 0.0 or resolution and abs(value) < resolution:
        return 0
    dec = get_decades(value)
    low = min(dec, 0)
    high = MAX_DIGITS
    d = d0 = abs(value)
    while high - low > 1:
        p = round(0.5 * (high + low))
        v = round(value, p)
        d = max(d0, abs(v))
        if abs(value - v) > reltol * d:
            low = p
        else:
            high = p
    v = round(value, low)
    if abs(value - v) < reltol * d:
        return low
    else:
        return high


def get_decades(value: int | float) -> int:
    """Compute number of digits needed to represent integer part of 'value' in fixed format.

    Args:
        value: ... to check

    Returns:
        Number of decades.
    """
    if value != 0:
        decimal_power = np.log10(abs(value))
    else:
        decimal_power = 0
    decades = np.trunc(decimal_power)
    if decimal_power < 0:
        decades -= 1
    return int(decades)


def significant_array(
    array: ndarray, reltol: float = FLOAT_TOLERANCE, resolution: float = None
) -> ndarray:
    """Compute the minimum numbers of significant digits to achieve desired tolerance."""
    result: ndarray = np.empty_like(array, dtype=int)
    for index in zip(*map(np.ravel, np.indices(array.shape))):
        result[index] = significant_digits(array[index], reltol, resolution)
    return result


def round_scalar(value: float, digits: int = None) -> float:
    """Rounds scalar value to represent the value in minimal form.

    Args:
        value: The value to be rounded.
        digits: The number of significant digits, optional.

    Returns:
        Rounded value.
    """
    if digits is None:
        digits = significant_digits(value, FLOAT_TOLERANCE, FLOAT_TOLERANCE)
    return round(value, digits)


def round_array(array: ndarray, digits_array: ndarray = None) -> ndarray:
    """Rounds array to desired precision.

    Args:
        array:   Array of values.
        digits_array:   Array of corresponding significant digits.

    Returns:
        Rounded array.
    """
    if digits_array is None:
        digits_array = significant_array(array, FLOAT_TOLERANCE, FLOAT_TOLERANCE)
    result: ndarray = np.empty_like(array)
    for index in zip(*map(np.ravel, np.indices(array.shape))):
        result[index] = round_scalar(array[index], digits_array[index])
    return result


@functools.singledispatch
def are_equal(a, b) -> bool:
    """Check if objects are equal dispatching method."""
    return a is b or a == b


@are_equal.register
def _(a: str, b: str) -> bool:
    return a is b or a == b


@are_equal.register
def _(a: ndarray, b: ndarray) -> bool:
    return np.array_equal(a, b)


@are_equal.register
def _(a: collections.abc.Iterable, b) -> bool:
    if not issubclass(type(b), collections.abc.Iterable):
        return False
    return all(are_equal(ai, bi) for ai, bi in itertools.zip_longest(a, b))


@functools.singledispatch
def is_in(where, x) -> bool:
    """Check if 'x' belongs to 'where' dispatcher."""
    if where is None:
        return False
    return x is where or x == where


@is_in.register
def _(where: str, x) -> bool:
    return x is where or x == where


@is_in.register
def _(where: tuple, x) -> bool:
    return any(is_in(i, x) for i in where)


@is_in.register
def _(where: collections.abc.Callable, x) -> bool:
    return where(x)  # type: ignore


@is_in.register
def _(where: collections.abc.Container, x) -> bool:
    return x in where


def filter_dict(
    a: Dict[Any, Any],
    *drop_items,
) -> Dict[Any, Any]:
    """Create copy of a dictionary omitting some keys."""
    res = {}
    for k, v in a.items():
        if drop_items and is_in(drop_items, k):
            pass
        else:
            if isinstance(v, dict):
                res[k] = filter_dict(v, *drop_items)
            elif issubclass(type(v), collections.abc.Collection):
                res[k] = deepcopy(v)
            else:
                res[k] = v
    return res


@functools.singledispatch
def make_hashable(x):
    """Create hashable object from 'x'."""
    raise TypeError(f"Don't know how to make {type(x).__name__} objects hashable")


@make_hashable.register
def _(x: collections.abc.Hashable):
    return x


@make_hashable.register
def _(x: str):
    return x


@make_hashable.register
def _(x: collections.abc.Mapping) -> Tuple:
    return tuple((k, make_hashable(v)) for k, v in x.items())


@make_hashable.register
def _(x: collections.abc.Iterable) -> Tuple:
    return tuple(map(make_hashable, x))


def compute_hash(*items) -> int:
    """Compute hash for a sequence of potentially formally not hashable objects.

    Note:
        Take care on the objects values are stable, while the hashes are in use.
    """
    if 1 < len(items):
        return compute_hash(tuple(map(compute_hash, items)))
    return hash(make_hashable(items[0]))


def is_sorted(a: np.ndarray) -> bool:
    """Check if an array is sorted."""
    return cast(bool, np.all(np.diff(a) > 0))


def mids(a: np.ndarray) -> np.ndarray:
    """Get centers of bins presented with array 'a'."""
    result: ndarray = 0.5 * (a[1:] + a[:-1])
    return result


def prettify_float(x: float, fmt: str = "{:.13g}") -> str:
    """Format float in uniform way."""
    if x.is_integer():
        return str(int(x))
    else:
        return fmt.format(x)



from typing import List, Optional, Tuple

import mckit.parser.common.utils as pu  # parse utils
import sly

from mckit.parser.common import Lexer as LexerBase
from mckit.parser.common import TransformationStrictIndex
from mckit.parser.common.utils import drop_c_comments, extract_comments
from mckit.surface import Surface, create_surface
from mckit.utils.indexes import Index

SURFACE_TYPES = {
    "P",
    "PX",
    "PY",
    "PZ",
    "S",
    "SO",
    "SX",
    "SY",
    "SZ",
    "CX",
    "CY",
    "CZ",
    "C/X",
    "C/Y",
    "C/Z",
    "KX",
    "KY",
    "KZ",
    "K/X",
    "K/Y",
    "K/Z",
    "TX",
    "TY",
    "TZ",
    "SQ",
    "GQ",
    "X",
    "Y",
    "Z",
    "RPP",
    "RCC",
    "BOX",
}


def intern_surface_type(word: str):
    word = pu.ensure_upper(word)
    word, res = pu.internalize(word, SURFACE_TYPES)
    if not res:
        raise pu.ParseError(f"{word} is not a valid surface type")
    return word


# noinspection PyPep8Naming,PyUnboundLocalVariable,PyUnresolvedReferences,SpellCheckingInspection
class Lexer(LexerBase):
    tokens = {MODIFIER, SURFACE_TYPE, FLOAT, INTEGER}

    MODIFIER = r"^\s{,5}(\*|\+)"
    SURFACE_TYPE = r"[a-z]+(?:/[a-z]+)?"
    FLOAT = pu.FLOAT
    INTEGER = pu.INTEGER

    @_(r"[a-z]+(?:/[a-z]+)?")
    def SURFACE_TYPE(self, t):
        t.value = intern_surface_type(t.value)
        return t

    @_(pu.FLOAT)
    def FLOAT(self, t):
        return LexerBase.on_float(t, use_zero=False)

    @_(pu.INTEGER)
    def INTEGER(self, t):
        return LexerBase.on_integer(t, use_zero=False)


# noinspection PyUnresolvedReferences
class Parser(sly.Parser):
    tokens = Lexer.tokens

    def __init__(self, transformations: Index):
        sly.Parser.__init__(self)
        self._transformations = transformations

    @property
    def transformations(self):
        return self._transformations

    def build_surface(
        self, name: int, kind: str, params: List[float], transform, modifier
    ) -> Surface:
        options = {"name": name}
        if transform is not None:
            transformation = self.transformations[transform]
            if transformation:
                options["transform"] = transformation
        if modifier is not None:
            options["modifier"] = modifier
        _surface = create_surface(kind, *params, **options)
        return _surface

    @_("MODIFIER  name surface_description")
    def surface(self, p):
        (
            kind,
            params,
            transform,
        ) = p.surface_description  # type: str, List[float], Optional[int]
        return self.build_surface(p.name, kind, params, transform, p.MODIFIER)

    @_("name surface_description")
    def surface(self, p):
        (
            kind,
            params,
            transform,
        ) = p.surface_description  # type: str, List[float], Optional[int]
        return self.build_surface(p.name, kind, params, transform, None)

    @_("INTEGER")
    def name(self, p):
        return p.INTEGER

    @_("transform SURFACE_TYPE surface_params")
    def surface_description(self, p) -> Tuple[str, List[float], Optional[int]]:
        return p.SURFACE_TYPE, p.surface_params, p.transform

    @_("SURFACE_TYPE surface_params")
    def surface_description(self, p) -> Tuple[str, List[float], Optional[int]]:
        return p.SURFACE_TYPE, p.surface_params, None

    @_("INTEGER")
    def transform(self, p):
        return p.INTEGER

    @_("surface_params float")
    def surface_params(self, p) -> List[float]:
        surface_params: List[float] = p.surface_params
        surface_params.append(p.float)
        return surface_params

    @_("float")
    def surface_params(self, p) -> List[float]:
        return [p.float]

    @_("FLOAT")
    def float(self, p):
        return p.FLOAT

    @_("INTEGER")
    def float(self, p) -> float:
        return float(p.INTEGER)


def parse(text: str, transformations: Optional[Index] = None) -> Surface:
    if transformations is None:
        transformations = TransformationStrictIndex()
    else:
        assert isinstance(transformations, Index)
    text = drop_c_comments(text)
    text, comments, trailing_comments = extract_comments(text)
    lexer = Lexer()
    parser = Parser(transformations)
    result = parser.parse(lexer.tokenize(text))
    if trailing_comments:
        result.options["comment"] = trailing_comments
    return result



"""Tests with benchmarks on large mcnp files.

To use it install plugin pytest-benchmark (https://pytest-benchmark.readthedocs.io/en/latest/index.html#)
    conda install pytest-benchmark
    or
    pip install pytest-benchmark
"""
from zipfile import ZipFile

import pytest

from mckit import Universe
from mckit.constants import MCNP_ENCODING
from mckit.parser.mcnp_input_parser import read_mcnp_text
from mckit.parser.mcnp_input_sly_parser import ParseResult, from_text
from mckit.utils.resource import path_resolver

data_filename_resolver = path_resolver("benchmarks")
with ZipFile(data_filename_resolver("data/4M.zip")) as data_archive:
    CLITE_TEXT = data_archive.read("clite.i").decode(encoding=MCNP_ENCODING)


def test_old_mcnp_reading(benchmark):
    universe: Universe = benchmark(read_mcnp_text, CLITE_TEXT)
    assert len(universe) == 150


def test_sly_mcnp_reading(benchmark):
    result: ParseResult = benchmark(from_text, CLITE_TEXT)
    assert result.title == "C-LITE VERSION 1 RELEASE 131031 ISSUED 31/10/2013 - Halloween edition"
    assert len(result.universe) == 150


if __name__ == "__main__":
    pytest.main()



import numpy as np

# noinspection PyUnresolvedReferences,PyPackageRequirements
from mckit.geometry import EX, EY, EZ
from mckit.geometry import GLOBAL_BOX as _GLOBAL_BOX
from mckit.geometry import Box as _Box


class Box(_Box):
    __doc__ = _Box.__doc__

    def __init__(self, center, wx, wy, wz, ex=EX, ey=EY, ez=EZ):
        _Box.__init__(self, center, wx, wy, wz, ex=ex, ey=ey, ez=ez)

    @classmethod
    def from_geometry_box(cls, geometry_box: _Box):
        return cls(
            geometry_box.center,
            *geometry_box.dimensions,
            geometry_box.ex,
            geometry_box.ey,
            geometry_box.ez,
        )

    def __eq__(self, other):
        return (
            np.array_equal(self.center, other.center)
            and np.array_equal(self.dimensions, other.dimensions)
            and np.array_equal(self.ex, other.ex)
            and np.array_equal(self.ey, other.ey)
            and np.array_equal(self.ez, other.ez)
        )

    def __hash__(self):
        return hash((self.center, self.dimensions, self.ex, self.ey, self.ez))

    def __getstate__(self):
        center = self.center
        wx, wy, wz = self.dimensions
        ex = self.ex
        ey = self.ey
        ez = self.ez
        return center, wx, wy, wz, ex, ey, ez

    def __setstate__(self, state):
        center, wx, wy, wz, ex, ey, ez = state
        self.__init__(center, wx, wy, wz, ex, ey, ez)


GLOBAL_BOX = Box.from_geometry_box(_GLOBAL_BOX)



import re

import mckit.parser.common as cmn
import pytest
import sly

from mckit.parser.common.Lexer import Lexer as LexerBase
from mckit.parser.common.Lexer import LexError


# noinspection PyUnboundLocalVariable,PyPep8Naming,PyUnresolvedReferences
class DerivedLexer(LexerBase):
    tokens = {FRACTION, FLOAT, INTEGER, ZERO}

    FRACTION = r"\d+(?:\.\d+[a-z])"

    @_(cmn.FLOAT)
    def FLOAT(self, token):
        return self.on_float(token)

    @_(cmn.INTEGER)
    def INTEGER(self, token):
        return self.on_float(token)

    ZERO = r"0"


@pytest.mark.parametrize(
    "text, expected_types, expected_values",
    [
        ("1 0 3.14", ["INTEGER", "ZERO", "FLOAT"], [1, 0, 3.14]),
        ("3.14 3.14c", ["FLOAT", "FRACTION"], [3.14, "3.14c"]),
        ("1 0 1e-4", ["INTEGER", "ZERO", "FLOAT"], [1, 0, 1e-4]),
    ],
)
def test_derived_lexer(text, expected_types, expected_values):
    lexer = DerivedLexer()
    tokens = list(lexer.tokenize(text))
    result = list(t.type for t in tokens)
    assert result == expected_types
    result = list(t.value for t in tokens)
    assert result == expected_values


@pytest.mark.parametrize(
    "text, msg_contains",
    [
        ("1~ 0 3.14", "column 2"),
        ("\n1 0 3.14 ~", "at line 2, column 10"),
        ("\n1 0 3.14 ~", r"\s{9}\^"),
        ("\n1 0 3.14 0~", r"\s{10}\^"),
        ("~", "column 1\n~\n\\^"),
    ],
)
def test_bad_path(text, msg_contains):
    lexer = DerivedLexer()
    with pytest.raises(LexError, match=msg_contains):
        for _ in lexer.tokenize(text):
            pass


# noinspection PyUnboundLocalVariable,PyPep8Naming,PyUnresolvedReferences
class MyLexer(LexerBase):
    literals = {":", "(", ")"}
    ignore = " \t"
    reflags = re.IGNORECASE | re.MULTILINE

    tokens = {NAME, FLOAT, INTEGER, ZERO}

    NAME = r"\d?[A-Za-z-]+"

    @_(cmn.FLOAT)
    def FLOAT(self, token):
        return self.on_float(token)

    @_(cmn.INTEGER)
    def INTEGER(self, token):
        return self.on_integer(token)

    ZERO = r"0"


@pytest.mark.parametrize(
    "text, expected_types, expected_values",
    [
        ("AAA 1 0 3.14", ["NAME", "INTEGER", "ZERO", "FLOAT"], ["AAA", 1, 0, 3.14]),
        ("1B 1 0 3.14", ["NAME", "INTEGER", "ZERO", "FLOAT"], ["1B", 1, 0, 3.14]),
    ],
)
def test_good_path(text, expected_types, expected_values):
    lexer = MyLexer()
    tokens = list(lexer.tokenize(text))
    result = list(t.type for t in tokens)
    assert result == expected_types
    result = list(t.value for t in tokens)
    assert result == expected_values


# noinspection PyUnresolvedReferences
class MyParser(sly.Parser):
    tokens = MyLexer.tokens

    @_("NAME number parameters")
    def expression(self, p):
        return p.NAME, p.number, p.parameters

    @_("NAME ZERO")
    def expression(self, p):
        return p.NAME, 0, None

    @_("INTEGER")
    def number(self, p):
        return p.INTEGER

    @_("parameters FLOAT")
    def parameters(self, p):
        res = p.parameters
        res.append(p.FLOAT)
        return res

    @_("FLOAT")
    def parameters(self, p):
        return [p.FLOAT]


@pytest.mark.parametrize(
    "text, expected",
    [("AAA 1 1.2 3.4", ("AAA", 1, [1.2, 3.4])), ("A-Z 0", ("A-Z", 0, None))],
)
def test_parser_with_derived_lexer(text, expected):
    lexer = MyLexer()
    parser = MyParser()
    actual = parser.parse(lexer.tokenize(text))
    assert actual == expected


if __name__ == "__main__":
    pytest.main()
